<?php

include 'class-cybs-soapi-cc.php';

if ( ! defined( 'ABSPATH' ) ) {
	exit; // Exit if accessed directly
}

/**
 * Credit Card Payments for Woocommerce with CyberSource implementation class.
 *
 * @class 		CyberSource_CC_Gateway
 * @extends		WC_Payment_Gateway
 * @version		1.0.0
 */
class CyberSource_CC_Gateway extends WC_Payment_Gateway {

    private $soap_client;
    

	/**
	 * Constructor.
	 */
	public function __construct() {
		$this->id                 = 'cybersource_cc';
		$this->method_title       = __( 'MoreFlo Pay', 'moreflo-checkout' );
		$this->method_description = __( 'Take credit card payments and use tokens for cards on file using the CyberSource platform. Requires SSL when in live mode. Request a test account here: <a href="https://www.cybersource.com/about/contact_us/" target="_blank">CyberSource</a>', 'moreflo-checkout' );		$this->has_fields         = true;
        $this->supports           = array(
            'products',
            'tokenization',
        );

        // Initialize the logger
        $this->logger = wc_get_logger();
        $this->logger_context = array( 'source' => $this->id );

		// Load the form fields
		$this->init_form_fields();

		// Load the settings.
		$this->init_settings();

		// Get setting values
        $this->enabled         = $this->get_option( 'enabled' );
        $this->sandbox         = $this->get_option( 'sandbox' );
        $this->card_on_file    = $this->get_option( 'card_on_file' );
        $this->microform       = $this->get_option( 'microform' );
        $this->title           = $this->get_option( 'title' );
		$this->description     = $this->get_option( 'description' );
		$this->merchantID      = $this->sandbox == 'no' ? $this->get_option( 'live_merchant_id' ) : $this->get_option( 'sandbox_merchant_id' );
        $this->soap_key        = $this->sandbox == 'no' ? $this->get_option( 'live_soap_key' ) : $this->get_option( 'sandbox_soap_key' );
        $this->shared_secret_key = $this->sandbox == 'no' ? $this->get_option( 'live_shared_secret_key' ) : $this->get_option( 'sandbox_shared_secret_key' );
        $this->shared_secret_key_serial_number = $this->sandbox == 'no' ? $this->get_option( 'live_shared_secret_key_serial_number' ) : $this->get_option( 'sandbox_shared_secret_key_serial_number' );
        $this->wsdl_version    = $this->get_option( 'wsdl_version' );
        $this->log_enabled     = $this->get_option( 'log_enabled' );

		// General Hooks
        add_action( 'woocommerce_update_options_payment_gateways_' . $this->id, array( $this, 'process_admin_options' ) );
        add_action( 'wp_enqueue_scripts', array( $this, 'payment_scripts' ) );      
        
	}
    public function payment_scripts() {
 
        // we need JavaScript to process a token only on checkout pages, right?
        if ( ! is_checkout() ) {
            return;
        }
     
        // if our payment gateway is disabled, we do not have to enqueue JS too
        if ( 'no' === $this->enabled ) {
            return;
        }
     
        wp_register_script( 'moreflow_pay', plugins_url( 'js/moreflow_Pay.js', __FILE__ ), array( 'jquery') );
        wp_register_style( 'moreflow_pay_css', plugins_url( 'css/moreflow_Pay.css' , __FILE__) );
        wp_enqueue_script( 'moreflow_pay' );
        wp_enqueue_style( 'moreflow_pay_css' );
     
    }

    public function validate_fields(){//Custom error messages
        if( empty( $_POST[ 'cybersource_cc-card-number' ]) ) {
            wc_add_notice(  'Card number is required!', 'error' );
            return false;
        }
        if( empty( $_POST[ 'cybersource_cc-card-expiry' ]) ) {
            wc_add_notice(  'Card expiry date is required!', 'error' );
            return false;
        }
        if( empty( $_POST[ 'cybersource_cc-card-cvc' ]) ) {
            wc_add_notice(  'CVC is required!', 'error' );
            return false;
        }
        return true;
    }

	/**
	 * Initialise Gateway Settings Form Fields.
	 */
	public function init_form_fields() {

        $label = __( 'Enable Logging', 'moreflo-checkout' );
        $description = __( 'Enable the logging of errors.', 'moreflo-checkout' );
        
        if ( defined( 'WC_LOG_DIR' ) ) {
            $log_url = add_query_arg( 'tab', 'logs', add_query_arg( 'page', 'wc-status', admin_url( 'admin.php' ) ) );
            $log_key = 'cybersource_cc-' . sanitize_file_name( wp_hash( 'cybersource_cc' ) ) . '-log';
            $log_url = add_query_arg( 'log_file', $log_key, $log_url );
        
            $label .= ' | ' . sprintf( __( '%1$sView Log%2$s', 'moreflo-checkout' ), '<a href="' . esc_url( $log_url ) . '">', '</a>' );
        }

		$this->form_fields = array(
			'enabled' => array(
				'title'       => __( 'Enable/Disable', 'moreflo-checkout' ),
				'label'       => __( 'Enable Credit Cards payment processing and tokenization using CyberSource', 'moreflo-checkout' ),
				'type'        => 'checkbox',
				'description' => '',
                'default'     => 'no'
            ),
            'sandbox' => array(
				'title'       => __( 'Sandbox', 'moreflo-checkout' ),
				'label'       => __( 'Enable Sandbox Mode', 'moreflo-checkout' ),
				'type'        => 'checkbox',
				'description' => __( 'Enabling test mode(real payments will not be taken).', 'moreflo-checkout' ),
				'default'     => 'yes'
            ),
            'card_on_file' => array(
				'title'       => __( 'Card On File', 'moreflo-checkout' ),
				'label'       => __( 'Enable Card On File.', 'moreflo-checkout' ),
				'type'        => 'checkbox',
				'description' => __( 'This allows customers to save cards to thier account. It stores the card info at CyberSource and a token representing the card is stored in your database.', 'moreflo-checkout' ),
                'default'     => 'no'
            ),
			'title' => array(
				'title'       => __( 'Title', 'moreflo-checkout' ),
				'type'        => 'text',
				'description' => __( 'This controls the title which the user sees during checkout within the credit card payment type.', 'moreflo-checkout' ),
				'default'     => __( 'Pay with Credit Card ', 'moreflo-checkout' )
			),
			'description' => array(
				'title'       => __( 'Description', 'moreflo-checkout' ),
				'type'        => 'text',
				'description' => __( 'This controls the description which the user sees during checkout within the credit card payment type.', 'moreflo-checkout' ),
				'default'     => __( 'Pay with your Credit Card via Cybersource payment gateway.', 'moreflo-checkout' )
			),
			'sandbox_merchant_id' => array(
				'title'       => __( 'Sandbox - Merchant ID', 'moreflo-checkout' ),
				'type'        => 'text',
				'description' => __( 'This is the sandbox account merchant ID', 'moreflo-checkout' ),
				'default'     => ''
			),
			'sandbox_soap_key' => array(
				'title'       => __( 'Sandbox - SOAP Key', 'moreflo-checkout' ),
				'type'        => 'text',
				'description' => __( 'This key is required for basic credit card processing in the sandbox environment. Get your SOAP API key from the <a href="https://ebc2test.cybersource.com" target="_blank">test CyberSource Enterprise Business Center</a>.', 'moreflo-checkout' ),
				'default'     => ''
            ),
			'live_merchant_id' => array(
				'title'       => __( 'Live - Merchant ID', 'moreflo-checkout' ),
				'type'        => 'text',
				'description' => __( 'This is the live account merchant ID', 'moreflo-checkout' ),
				'default'     => ''
			),
			'live_soap_key' => array(
				'title'       => __( 'Live - SOAP Key', 'moreflo-checkout' ),
				'type'        => 'text',
				'description' => __( 'This key is required for basic credit card processing in the production environment. Get your SOAP API key from the <a href="https://ebc2.cybersource.com" target="_blank">production CyberSource Enterprise Business Center</a>.', 'moreflo-checkout' ),
				'default'     => ''
            ),
            'wsdl_version' => array(
                'title'       => __( 'WSDL Version', 'moreflo-checkout' ),
                'type'        => 'text',
                'description' => __( 'The version number of the CyberSource WSDL. Used by the SOAP API.', 'moreflo-checkout' ),
                'default'     => '149'
            ),
            'log_enabled' => array(
                'title'       => __( 'Debug Log', 'moreflo-checkout' ),
                'label'       => $label,
                'description' => $description,
                'type'        => 'checkbox',
                'default'     => 'no'
            )
		);
    }
    
    /**
     * Check if the gateway is enabled and correctly setup.
     *
     * @return bool
     */
    public function is_available() {

        // Gateway itself must be enabled
        if ( 'no' == $this->enabled ) {
            return false;
        }

        // If not in sandbox mode HTTPS must be used
        if ( 'no' == $this->sandbox && ! wc_checkout_is_https() ) {
            if ( $this->log_enabled == 'yes' ) {
                $this->logger->alert("Live mode enabled, but site is not HTTPS", $this->logger_context);
            }
            return false;
        }

        // Merchant ID, key, and WSDL version must be set
        if ( ! $this->merchantID || ! $this->soap_key || ! $this->wsdl_version) {
            if ( $this->log_enabled == 'yes' ) {
                $this->logger->alert("Merchant ID, key, or WSDL version not set in payment gateway configuration", $this->logger_context);
            }
            return false;
        }

        return true;
    }

	/**
	 * Payment form on checkout page.
	 */
	public function payment_fields() {
		$description = $this->get_description();

		if ( 'yes' == $this->sandbox ) {
			$description .= ' ' . 'Test Card: 4111111111111111' ;
		}

		if ( $description ) {
			echo wpautop( wptexturize( trim( $description ) ) );
		}

        parent::payment_fields();
	}

	/**
	 * Process the payment.
	 *
	 * @param int $order_id
	 */
	public function process_payment( $order_id ) {
        $save_card_on_file = false;
        $tokenReasonCode = "";
        
        $order = new WC_Order( $order_id );

        if ( $this->log_enabled == 'yes' ) {
            $this->logger->debug("****** Posted Data from checkout page:", $this->logger_context);
            $this->logger->debug(wc_print_r($_POST, true), $this->logger_context);
        }

        // Create a Cybs Client and set all the fields in the request
        $this->soap_client = new CybsSoapiCC($this->merchantID, $this->soap_key, $this->wsdl_version, $this->sandbox);
        $request = $this->soap_client->createRequest($order_id);
        $this->soap_client->setAuthInRequest($request);

        // Request capture of funds if immediate capture/settlement enabled.
        $this->soap_client->setCaptureInRequest($request);

        $this->setBillToInRequest($request, $order);
        //$this->setShipToInRequest($request, $order);
        $this->soap_client->setPurchaseTotals($request, $order->get_currency(), $order->get_total());

        // Saving this payment method as a token?
        if ( is_user_logged_in() && isset( $_POST['wc-cybersource_cc-new-payment-method'] ) && true === (bool) $_POST['wc-cybersource_cc-new-payment-method'] ) {
            
            $save_card_on_file = true;
            
            if ( 'yes' == $this->microform ) {
                $subscriptionID = $_POST['microform_token'];
            } else {
                $this->soap_client->setSubscriptionCreateInRequest($request);
            }
        }

        // Using a card on file to process the payment?
        if ( is_user_logged_in() && isset( $_POST['wc-cybersource_cc-payment-token'] ) && 'new' !== $_POST['wc-cybersource_cc-payment-token'] ) {
            $token_id = $_POST['wc-cybersource_cc-payment-token'];
            $token = WC_Payment_Tokens::get( $token_id );

            //Verify token belongs to the logged in user
            if ( $token->get_user_id() !== get_current_user_id() ) {
                wc_add_notice( __('Payment error:', 'woothemes') . 'Payment Failed. User not associated to token ', 'error' );
                return;
            }

            $this->soap_client->setTokenInRequest($request, $token->get_token());
        // Using the Microform?
        } else if ( 'yes' == $this->microform ) {
            $this->soap_client->setTokenInRequest($request, $_POST['microform_token']);
        // Not using Microform or a card on file?
        } else {
            $this->setCardInRequest($request);
        }

        if ( $this->log_enabled == 'yes' ) {
            $this->logger->debug("****** Request To CyberSource for Payment:", $this->logger_context);
            $this->logger->debug(wc_print_r($request, true), $this->logger_context);
        }

        // Use the Cybs soap_client to execute the transaction
        $reply = $this->soap_client->runTransaction($request);

        if ( $this->log_enabled == 'yes' ) {
            $this->logger->debug("****** Response from CyberSource for Payment:", $this->logger_context);
            $this->logger->debug(wc_print_r($reply, true), $this->logger_context);
        }

        $decision = $reply->decision;
        $reasonCode = $reply->reasonCode;

        if(isset ($reply->ccAuthReply)) {
            $authReasonCode = $reply->ccAuthReply->reasonCode;
        }

        if(isset ($reply->paySubscriptionCreateReply)) {
            $tokenReasonCode = $reply->paySubscriptionCreateReply->reasonCode;
            $subscriptionID = $reply->paySubscriptionCreateReply->subscriptionID;
        }

        if(strcmp($decision,'ACCEPT') == 0 && strcmp($reasonCode, '100') == 0) {
            // Remove cart
            WC()->cart->empty_cart();

            // Complete the order
            $order->payment_complete();

            // If we are saving this card to file and either a token was returned in the reply or we are using microform
            if($save_card_on_file && (strcmp($tokenReasonCode,'100') == 0 || 'yes' == $this->microform)) {
                $this->saveToken($subscriptionID);
            }

            // Return thank you redirect
            return array(
                'result' => 'success',
                'redirect' => $this->get_return_url( $order )
            );

        } else {
            wc_add_notice( __('Payment error:', 'woothemes') . 'Payment Failed. Reason Code: ' . $reply->reasonCode, 'error' );
            return;
        }
	}

	/**
	 * Get gateway icon.
	 *
	 * @access public
	 * @return string
	 */
	public function get_icon() {
		$icon  = '<p class="icon-sec"><img src="' . WC_HTTPS::force_https_url( WC()->plugin_url() . '/assets/images/icons/credit-cards/visa.png' ) . '" alt="Visa" />';
		$icon .= '<img src="' . WC_HTTPS::force_https_url( WC()->plugin_url() . '/assets/images/icons/credit-cards/mastercard.png' ) . '" alt="MasterCard" />';
		$icon .= '<img src="' . WC_HTTPS::force_https_url( WC()->plugin_url() . '/assets/images/icons/credit-cards/discover.png' ) . '" alt="Discover" />';
		$icon .= '<img src="' . WC_HTTPS::force_https_url( WC()->plugin_url() . '/assets/images/icons/credit-cards/amex.png' ) . '" alt="Amex" />';
		$icon .= '<img src="' . WC_HTTPS::force_https_url( WC()->plugin_url() . '/assets/images/icons/credit-cards/jcb.png' ) . '" alt="JCB" /></p>';

		return apply_filters( 'woocommerce_gateway_icon', $icon, $this->id );
	}

	public function field_name( $name ) {
		return ' name="' . esc_attr( $this->id . '-' . $name ) . '" ';
	}

    /**
     * Add a Credit Card as a Payment Method. This tokenizes and stores
     * the token for use at checkout.
     *
     * @access public
     */
    public function add_payment_method() {

        if(! is_user_logged_in()) {
            wc_add_notice( __( 'User not logged in, cannot add card.', 'moreflo-checkout' ), 'error' );
            return;
        }

        if ( $this->log_enabled == 'yes' ) {
            $this->logger->debug("****** Posted Data from account management page:", $this->logger_context);
            $this->logger->debug(wc_print_r($_POST, true), $this->logger_context);
        }

        // Create a Cybs Client and set all the fields in the request
        $this->soap_client = new CybsSoapiCC($this->merchantID, $this->soap_key, $this->wsdl_version, $this->sandbox);
        $request = $this->soap_client->createRequest('12345');
        $this->setCardInRequest($request);

        $this->soap_client->setSubscriptionCreateInRequest($request);

        $purchaseTotals = new stdClass();
        $purchaseTotals->currency = 'USD';
        $request->purchaseTotals = $purchaseTotals;

        global $woocommerce;
        $this->setBillToInRequest($request, $woocommerce->customer);

        if ( $this->log_enabled == 'yes' ) {
            $this->logger->debug("****** Request to CyberSource to tokenize:", $this->logger_context);
            $this->logger->debug(wc_print_r($request, true), $this->logger_context);
        }

        // Use the Cybs soap_client to execute the transaction
        $reply = $this->soap_client->runTransaction($request);

        if ( $this->log_enabled == 'yes' ) {
            $this->logger->debug("****** Response from CyberSource to tokenize:", $this->logger_context);
            $this->logger->debug(wc_print_r($reply, true), $this->logger_context);
        }

        $decision = $reply->decision;
        $reasonCode = $reply->reasonCode;

        if(isset ($reply->paySubscriptionCreateReply)) {
            $tokenReasonCode = $reply->paySubscriptionCreateReply->reasonCode;
            $subscriptionID = $reply->paySubscriptionCreateReply->subscriptionID;
        }

        if(strcmp($decision,'ACCEPT') == 0 && strcmp($reasonCode, '100') == 0) {

            // If tokenization successful save the token
            if(strcmp($tokenReasonCode,'100') == 0) {
                $this->saveToken($subscriptionID);
            }

            // Return to payment methods page
            return array(
                'result'   => 'success',
                'redirect' => wc_get_endpoint_url( 'payment-methods' ),
            );

        } else {
            wc_add_notice( __( 'There was a problem adding this card.', 'moreflo-checkout' ), 'error' );
            return;
        }
    }

    /**
     * Saves a CyberSource token (subscriptionID) within a WooCommerce token.
     *
     * @access public
     * @param string $subscriptionID
     */
    public function saveToken($subscriptionID) {

        if ('yes' == $this->microform ) {
            $accountNumber = str_replace( array(' ', '-' ), '', $_POST['microform_masked_pan'] ); 
        } else {
            $accountNumber = str_replace( array(' ', '-' ), '', $_POST['cybersource_cc-card-number'] );
        }

        $first_six = substr($accountNumber,0,5);
        error_log("BIN : " . $first_six);

        $expirationDate = str_replace( array( '/', ' '), '', $_POST['cybersource_cc-card-expiry'] );
        $expirationMonth = substr($expirationDate,0,2);
        if(strlen($expirationDate) > 4) {
            $expirationYear = substr($expirationDate,2,4);
        } else {
            $expirationYear = substr($expirationDate,2,2);
        }

        $token = new WC_Payment_Token_CC();
        $token->set_token( $subscriptionID );
        $token->set_gateway_id( $this->id );
        $token->set_card_type( $this->soap_client->getCardType($first_six)[1] );
        $token->set_last4( substr($accountNumber,12,4) );
        $token->set_expiry_month( $expirationMonth );
        $token->set_expiry_year( '20' . $expirationYear );
        $token->set_user_id( get_current_user_id() );

        if(! $token->save()) {
            if ( $this->log_enabled == 'yes' ) {
                $this->logger->error("Failed to save token locally.", $this->logger_context);
            }
            wc_add_notice( __( 'There was a problem saving your card to file.', 'moreflo-checkout' ), 'error' );
        }

    }

    /**
     * Deletes a CyberSource token (subscriptionID) as well as the related WooCommerce token.
     *
     * @access public
     * @param string $subscriptionID
     */
    public function delete_payment_method($subscriptionID) {

        // Create a Cybs Client and set all the fields in the request
        $this->soap_client = new CybsSoapiCC($this->merchantID, $this->soap_key, $this->wsdl_version, $this->sandbox);
        $request = $this->soap_client->createRequest('12345');
        $this->soap_client->setSubscriptionDeleteInRequest($subscriptionID, $request);

        if ( $this->log_enabled == 'yes' ) {
            $this->logger->debug("Request to CyberSource to delete token.", $this->logger_context);
            $this->logger->debug(wc_print_r($request, true), $this->logger_context);
        }

        // Use the Cybs soap_client to execute the transaction
        $reply = $this->soap_client->runTransaction($request);

        if ( $this->log_enabled == 'yes' ) {
            $this->logger->debug("Response from CyberSource to delete token.", $this->logger_context);
            $this->logger->debug(wc_print_r($reply, true), $this->logger_context);
        }

        $decision = $reply->decision;
        $reasonCode = $reply->reasonCode;

        if(isset ($reply->paySubscriptionDeleteReply)) {
            $tokenReasonCode = $reply->paySubscriptionDeleteReply->reasonCode;
        }

    }
    // get user ip address
    // public function getUserIpAddr(){
    //     if(!empty($_SERVER['HTTP_CLIENT_IP'])){
    //         //ip from share internet
    //         $ip = $_SERVER['HTTP_CLIENT_IP'];
    //     }elseif(!empty($_SERVER['HTTP_X_FORWARDED_FOR'])){
    //         //ip pass from proxy
    //         $ip = $_SERVER['HTTP_X_FORWARDED_FOR'];
    //     }else{
    //         $ip = $_SERVER['REMOTE_ADDR'];
    //     }
    //     return $ip;
    // }
    public function setBillToInRequest($request, $order) {
        $billTo = new stdClass();
        $userIP = $_SERVER['REMOTE_ADDR'];
        $billTo->firstName = $order->get_billing_first_name();
        $billTo->lastName = $order->get_billing_last_name();
        $billTo->street1 = $order->get_billing_address_1();
        $billTo->city = $order->get_billing_city();
        $billTo->state = $order->get_billing_state();
        $billTo->postalCode = $order->get_billing_postcode();
        $billTo->country = $order->get_billing_country();
        $billTo->email = $order->get_billing_email();
        $billTo->ipAddress = $userIP;
        $request->billTo = $billTo;
    }

    public function setShipToInRequest($request, $order) {
        $shipTo = new stdClass();
        $shipTo->firstName = $order->get_shipping_first_name();
        $shipTo->lastName = $order->get_shipping_last_name();
        $shipTo->street1 = $order->get_shipping_address_1();
        $shipTo->city = $order->get_shipping_city();
        $shipTo->state = $order->get_shipping_state();
        $shipTo->postalCode = $order->get_shipping_postcode();
        $shipTo->country = $order->get_shipping_country();
        $request->shipTo = $shipTo;
    }

    public function setCardInRequest($request) {
        $card = new stdClass();
        $accountNumber = str_replace( array(' ', '-' ), '', $_POST['cybersource_cc-card-number'] );
        $card->accountNumber = $accountNumber;
        $expirationDate = str_replace( array( '/', ' '), '', $_POST['cybersource_cc-card-expiry'] );
        $card->expirationMonth = substr($expirationDate,0,2);
         if(strlen($expirationDate) > 4) {
             $card->expirationYear = '20'.substr($expirationDate,2,4);
         } else {
             $card->expirationYear = '20'.substr($expirationDate,2,2);
         }
        $request->card = $card;
    }
 
}
