<?php

if ( ! defined( 'ABSPATH' ) ) {
	exit; // Exit if accessed directly
}

/**
 * Credit Card Payments with Morflo
 *
 * @class 		NmiPayGateway
 * @extends		WC_Payment_Gateway
 * @version		1.0.0
 */
require_once(__DIR__ . '/vendor/autoload.php');
use TorneLIB\MODULE_CURL;
define('ORDER_TYPE', 'order_type');
define('DATETIME_FORMATE', "h:i:sa");
define('SUCCESS', 'success');
define('EKASHU_REFERENCE', 'ekashu_reference');
define('EKASHU_TRANSACTION_ID', 'ekashu_transaction_id');

class NmiPayGateway extends WC_Payment_Gateway {

    /**
     * Class constructor, more about it in Step 3
     */
    public $retry_interval;

	private $logger;
	private $logger_context;
	private $sandbox;
	private $terminal_id;
	private $transaction_key;
	private $hash_key;
	private $log_enabled;

    public function __construct() {
        $this->id                 = 'nmi_pay_gateway';
		$this->method_title       = __( 'MoreFlo Checkout', MOREFLO_CHECKOUT );
        $this->method_description = __( 'Use credit card payment with MoreFlo. Requires SSL in live mode.', MOREFLO_CHECKOUT );
        $this->has_fields         = false;
        $this->supports           = array(
            'products',
            'subscriptions',
            'subscription_cancellation',
            'subscription_suspension',
            'subscription_reactivation',
            'subscription_amount_changes',
            'subscription_date_changes',
            'subscription_payment_method_change',
            'subscription_payment_method_change_customer',
            'subscription_payment_method_change_admin',
            'multiple_subscriptions',
            'refunds',
        );

        $this->init_form_fields();
        $this->logger = wc_get_logger();
        $this->logger_context = array( 'source' => $this->id );

        // Get setting values
        $this->enabled         = $this->get_option( 'enabled' );
        $this->sandbox         = $this->get_option( 'sandbox' );
        $this->title           = $this->get_option( 'title' );
        $this->description     = $this->get_option( 'description' );
        $this->terminal_id     = $this->sandbox == 'no' ? $this->get_option( 'live_terminal_id' ) : $this->get_option( 'sandbox_terminal_id' );
        $this->transaction_key = $this->sandbox == 'no' ? $this->get_option( 'live_transaction_key' ) : $this->get_option( 'sandbox_transaction_key' );
        $this->hash_key        = $this->sandbox == 'no' ? $this->get_option( 'live_hash_key' ) : $this->get_option( 'sandbox_hash_key' );
        $this->log_enabled     = $this->get_option( 'log_enabled' );

        //Save settings fields
        add_action( 'woocommerce_update_options_payment_gateways_' . $this->id, array( $this, 'process_admin_options' ), 10, 0);

        add_action('woocommerce_api_nmipaygateway', array($this, 'nmi_callback_response'), 10, 0);

        // add_action( 'wp_enqueue_scripts', array( $this, 'payment_scripts' ), 10, 0 );

        add_action( 'woocommerce_scheduled_subscription_payment_'.$this->id, array( $this, 'scheduled_NMI_subscription_payment' ), 10, 2 );

    }

   /**
     * Plugin options, we deal with it in Step 3 too
     */
    public function init_form_fields(){
        $label = __( 'Enable Logging', MOREFLO_CHECKOUT );
        $description = __( 'Enable the logging of errors.', MOREFLO_CHECKOUT );

        if ( defined( 'WC_LOG_DIR' ) ) {
            $nmi_pay_log_url = add_query_arg( 'tab', 'logs', add_query_arg( 'page', 'wc-status', admin_url( 'admin.php' ) ) );
            $nmi_pay_log_key = 'nmi_pay-' . sanitize_file_name( wp_hash( 'nmi_pay' ) ) . '-log';
            $nmi_pay_log_url = add_query_arg( 'log_file', $nmi_pay_log_key, $nmi_pay_log_url );

            $label .= ' | ' . sprintf( __( '%1$sView Log%2$s', MOREFLO_CHECKOUT ), '<a href="' . esc_url( $nmi_pay_log_url ) . '">', '</a>' );
        }

		$this->form_fields = array(
            'enabled' => array(
				'title'       => __( 'Enable/Disable', MOREFLO_CHECKOUT ),
				'label'       => __( 'Enable credit card payment with MoreFlo', MOREFLO_CHECKOUT ),
				'type'        => 'checkbox',
				'description' => '',
                'default'     => 'no'
            ),
            'sandbox' => array(
				'title'       => __( 'Sandbox', MOREFLO_CHECKOUT ),
				'label'       => __( 'Enable Sandbox Mode', MOREFLO_CHECKOUT ),
				'type'        => 'checkbox',
				'description' => __( 'Enabling test mode(real payments will not be taken).', MOREFLO_CHECKOUT ),
				'default'     => 'yes'
            ),
			'title' => array(
				'title'       => __( 'title', MOREFLO_CHECKOUT ),
				'type'        => 'text',
				'description' => __( 'This controls the title which the user sees during checkout within the credit card payment type.', MOREFLO_CHECKOUT ),
				'default'     => __( 'Pay with credit card', MOREFLO_CHECKOUT )
			),
			'description' => array(
				'title'       => __( 'description', MOREFLO_CHECKOUT ),
				'type'        => 'text',
				'description' => __( 'This controls the description which the user sees during checkout within the credit card payment type.', MOREFLO_CHECKOUT ),
				'default'     => __( 'Pay with credit card.', MOREFLO_CHECKOUT )
			),
            'sandbox_terminal_id' => array(
				'title'       => __( 'Sandbox - Terminal-ID', MOREFLO_CHECKOUT ),
				'type'        => 'text',
				'description' => __( 'This is the sandbox account Terminal-ID', MOREFLO_CHECKOUT ),
				'default'     => ''
			),
			'sandbox_transaction_key' => array(
				'title'       => __( 'Sandbox Transaction Key', MOREFLO_CHECKOUT ),
				'type'        => 'text',
				'description' => __( 'This is the sandbox Transaction Key. Follow these instructions create accounts https://testwebmis.creditcall.com.', MOREFLO_CHECKOUT ),
				'default'     => ''
            ),
			'sandbox_hash_key' => array(
				'title'       => __( 'Sandbox Hash Key', MOREFLO_CHECKOUT ),
				'type'        => 'text',
				'description' => __( 'This is hash key for sandbox transaction. This hash key can be request from eKashu Support (support@ekashu.com)', MOREFLO_CHECKOUT ),
				'default'     => ''
            ),
			'live_terminal_id' => array(
				'title'       => __( 'Live - Terminal-ID', MOREFLO_CHECKOUT ),
				'type'        => 'text',
				'description' => __( 'This is the live account Terminal-ID.', MOREFLO_CHECKOUT ),
				'default'     => ''
			),
			'live_transaction_key' => array(
				'title'       => __( 'Live Transaction Key', MOREFLO_CHECKOUT ),
				'type'        => 'text',
				'description' => __( 'This is the live Transaction Key. Follow these instructions create accounts https://testwebmis.creditcall.com.', MOREFLO_CHECKOUT ),
				'default'     => ''
            ),
			'live_hash_key' => array(
				'title'       => __( 'Live Hash Key', MOREFLO_CHECKOUT ),
				'type'        => 'text',
				'description' => __( 'This is hash key for live transaction. This hash key can be request from eKashu Support (support@ekashu.com)', MOREFLO_CHECKOUT ),
				'default'     => ''
            ),
            'log_enabled' => array(
                'title'       => __( 'Debug Log', MOREFLO_CHECKOUT ),
                'label'       => $label,
                'description' => $description,
                'type'        => 'checkbox',
                'default'     => 'no'
            )
        );

    }

   /**
    * You will need it if you want your custom credit card form, Step 4 is about it
    */
   public function payment_fields() {
	if ( $this->description ) {
		echo wpautop( wp_kses_post( $this->description ) );
	}
   }

   /*
    * Custom CSS and JS, in most cases required only when you decided to go with a custom credit card form
    */
    public function payment_scripts() {

        wp_enqueue_style( 'moreflow_pay_nmi', plugin_dir_url(__FILE__). 'css/moreflow_pay_nmi.css', array(),'','' );

    }

   /*
     * Fields validation, more in Step 5
    */
   public function validate_fields() {



   }

   /*
    * We're processing the payments here, everything about it is in Step 5
    */
   public function process_payment( $order_id ) {

        global $woocommerce;

        $this->write_Woo_logs('process_payment');

        $nminonce =  wp_create_nonce('nmi_callback_mode');
        update_post_meta($order_id, '_nmi_payment_page', $nminonce);//Create security calls

        return array(
            'result' => SUCCESS,
            'redirect' => get_site_url().'/?wc-api=NmiPayGateway&event_type=proccess&order_id='.$order_id//get_site_url() sometime it might cause an issue for network path enable site
        );

    }

   /*
    * Callback methods
    */
    public function nmi_callback_response()
    {

        if($_REQUEST['event_type']){
            $event_type = $_REQUEST['event_type'];

            if(isset($_REQUEST['order_id'])){
                $order_id = sanitize_text_field($_REQUEST['order_id']);
            }

            $returnResult = empty( $_POST )
                ? false
                : $this->hashVerification( $_POST );

            //assign order id from post object from ekashu return on fail,success or callbacks methods
            if(isset($_POST[EKASHU_REFERENCE])){
                $order_id = $_POST[EKASHU_REFERENCE];
            }

            if ( $returnResult && 'success' == $_POST['ekashu_auth_result'] ) { //Success based hashed verification
                $this->successPayment($order_id, $event_type);
            } else {
                if($event_type == 'fail' || $event_type == 'fail_payment_nmi'){
                    $this->failPayment($order_id, $event_type);
                }
                if($event_type == 'proccess'){
                    $this->write_Woo_logs('proccess payment');
                    $this->processPayment($order_id);
                }
            }
        }

    }

    public function hashVerification($post)
    {
        if ( $post['ekashu_seller_id'] != $this->terminal_id ) {
            return false;
        }

        if ( $post['ekashu_seller_key'] != substr(str_replace(' ', '', $this->transaction_key),0,8) ) {
            return false;
        }

        $endpoint = ( $this->sandbox == 'yes' )
            ? 'https://test.ekashu.com'
            : 'https://live.ekashu.com';
        return NmiPayHashInput::validate_hash( $post, $endpoint );
    }

    private function processPayment($order_id)
    {

        $order = wc_get_order( $order_id );
        $metafield = get_post_meta($order_id, '_nmi_payment_page_proccess');

        if(count($metafield) > 0){
            wc_add_notice( __( 'This order already been proccess, Please try again later', MOREFLO_CHECKOUT ), 'error' );
            wp_safe_redirect(wc_get_checkout_url());
        }
        if($this->sandbox == "yes"){
            $endpoint = "https://test.ekashu.com";
        }else{
            $endpoint = "https://live.ekashu.com";
        }

        $ekashu_terminal_id = $this->terminal_id;
        $ekashu_transaction_key = substr(str_replace(' ', '', $this->transaction_key),0,8);//Only assigning first 8 digits

        // Get Order ID and Key
        $ekashu_id = $order_id;

        // Get Order Totals $0.00
        $ekashu_currency = $order->get_currency();
        $ekashu_total = $order->get_total();
        $woodecimaloptions = get_option('woocommerce_price_num_decimals');
        if($woodecimaloptions == 0){
            $ekashu_total = number_format((float)$ekashu_total, 2, '.', '');
        }else{
            $ekashu_total = $this->formateAmount($ekashu_total);
        }

        // Get Order User, Billing & Shipping Addresses
        $ekashu_email = $order->get_billing_email();
        $ekashu_phone = $order->get_billing_phone();

        // Get Callbacks, success and fail URLs
        $home = get_site_url();

        $ekashu_success_url = $home.'/?wc-api=NmiPayGateway&event_type=success_payment_nmi';
        $ekashu_failure_url = $home.'/?wc-api=NmiPayGateway&event_type=fail_payment_nmi';
        $ekashu_return_url = $order->get_cancel_order_url();
        $ekashu_callback_success_url = $home.'/?wc-api=NmiPayGateway&event_type=success';
        $ekashu_callback_failure_url = $home.'/?wc-api=NmiPayGateway&event_type=fail';

        $site_title = (get_bloginfo( 'name' ) != null)
            ? html_entity_decode( get_bloginfo( 'name' ) )
            : ' ';
        $site_description = (get_bloginfo( 'description' ) != null)
            ? html_entity_decode( get_bloginfo( 'description' ) )
            : ' ';

        $ekashu_pay_page_style = plugin_dir_url(__FILE__). 'css/moreflow_pay_nmi.css';
        $ekashu_shortcut_icon = plugin_dir_url(__FILE__).'images/northmill-fav.png';
        $ekashu_title = 'Northmill';

        $ekashu_request_type = 'auth';
        if(function_exists('wcs_order_contains_subscription')){
            $orderContainSubscription = wcs_order_contains_subscription( $order );
            if($orderContainSubscription === true && $ekashu_total == 0)
            {
                $ekashu_request_type = 'preauth';
            }
        }

        $hash_key = $this->hash_key;
        try {
            $hash_input = new NmiPayHashInput([
                'ekashu_amount' => $ekashu_total,
                'ekashu_currency' => $ekashu_currency,
                'ekashu_reference' => $ekashu_id,
                'ekashu_seller_id' => $ekashu_terminal_id,
                'ekashu_seller_key' => $ekashu_transaction_key,

                'ekashu_card_address_editable' => 'true',

                'ekashu_card_address_verify' => 'false',
                'ekashu_card_zip_code_verify' => 'false',
                'ekashu_card_title_mandatory' => 'false',
                'ekashu_delivery_title_mandatory' => 'false',

                'ekashu_seller_name' => $site_title,
                'ekashu_description' => $site_description,

                'ekashu_style_sheet' => $ekashu_pay_page_style,
                'ekashu_shortcut_icon' => $ekashu_shortcut_icon,
                'ekashu_title' => $ekashu_title,

                'ekashu_callback_include_post' => 'true',
                'ekashu_include_post' => 'true',
                'ekashu_auto_confirm' => 'true',

                'ekashu_viewport' => 'device-width, initial-scale=1',

                'ekashu_failure_url' => $ekashu_failure_url,
                'ekashu_success_url' => $ekashu_success_url,
                'ekashu_callback_success_url' => $ekashu_callback_success_url,
                'ekashu_callback_failure_url' => $ekashu_callback_failure_url,
                'ekashu_request_type' => $ekashu_request_type,
            ]);
        } catch ( Exception $e ) {
            wc_add_notice( __( 'Error during checkout. Please choose another payment option.', MOREFLO_CHECKOUT ), 'error' );
            wp_safe_redirect(wc_get_checkout_url());
            exit;
        }

        $computedHash = $hash_input->get_hash( $hash_key );

        echo '<p class="payinstruction">'.__( "Please wait! You are now redirected to a secure payment page", MOREFLO_CHECKOUT ).'</p>';
        echo '<form id="payment_page_form" action='.$endpoint.' method="post">';
        foreach ( $hash_input->get_form_data() as $key => $val ) {
            if ( is_null( $val ) ) {
                continue;
            }
            if ( is_bool( $val ) ) {
                echo '<input type="hidden" name="'. esc_attr( $key ) .'" value="'. ( $val ? 'true' : 'false' ) .'"/>';
            } else {
                echo '<input type="hidden" name="'. esc_attr( $key ) .'" value="'. esc_attr( $val ) .'"/>';
            }
        }
        echo '<input type="hidden" name="ekashu_hash_code" value="'. esc_attr( $computedHash ) .'"/>';
        echo '<input type="hidden" name="ekashu_return_url" value="'.$ekashu_return_url.'"/>';
        echo '<input type="hidden" name="ekashu_card_name_required" value="true"/>';
        echo '<input type="hidden" name="ekashu_card_name_mandatory" value="true"/>';
        echo '<input type="hidden" name="ekashu_card_phone_number" value="'. $ekashu_phone .'"/>';
        echo '<input type="hidden" name="ekashu_delivery_email_mandatory" value="false"/>';
        echo '<input type="hidden" name="ekashu_valid_from_required" value="false" />';
        echo '<input type="hidden" name="ekashu_card_email_address" value="'. $ekashu_email .'" />';
        echo '<input style="display:none" type="submit" value="Pay"/>';
        echo '<style type="text/css">#payment_page_form::before {height: 1em; width: 1em;display: block;position: absolute;
            top: 50%;left: 50%;margin-left: -0.5em;margin-top: -0.5em;content: "";animation: spin 1s ease-in-out infinite;
            background: url("/wp-content/plugins/moreflo-pay/images/loader.svg") center center;background-size: cover;line-height: 1;text-align: center;
            font-size: 2em;color: rgba(#000, 0.75)}@keyframes spin {from {transform:rotate(0deg);}to {transform:rotate(360deg);}}
            p.payinstruction{text-align:center;margin: 0 !important;}</style>';
        echo '<script type="text/javascript">
            document.getElementById("payment_page_form").submit();
            </script>
            </form>';
        update_post_meta($order_id, '_nmi_payment_page_proccess', 'hasproccess');//Create security calls
        wp_die();

    }

    private function failPayment($order_id, $event_type)
    {
        $order = new WC_Order($order_id);

        $current_order_status = $order->get_status();
        if($current_order_status != WC_FAILED_STATUS){//This is stop update again the status
            $this->write_Woo_logs('Fail payment');//Logs
            $order->update_status( WC_FAILED_STATUS );
        }

        if($event_type == 'fail_payment_nmi'){
            $this->write_Woo_logs($order_id.'- Fail payment redirect');//Logs
            wc_add_notice( __( 'Payment failed. Please contact bank for information', MOREFLO_CHECKOUT ), 'error' );
            wp_safe_redirect(wc_get_checkout_url());
        }

        if($event_type == 'fail'){//Callback fail method
            $this->write_Woo_logs($order_id.'- Fail callback response received');//Logs
            $returnResult[SUCCESS] = true;
            $returnResult['errorCode'] = "200";

            wp_send_json( $returnResult, $returnResult['errorCode'] );//Send received notification to server
        }


    }

    private function successPayment($order_id, $event_type)
    {
        $order = new WC_Order($order_id);

        update_post_meta($order_id, EKASHU_TRANSACTION_ID, $_POST[EKASHU_TRANSACTION_ID]);//Adding transaction id

        $this->subcriptions_order($order_id);

        if(isset($_POST[EKASHU_TRANSACTION_ID])){
            $order->set_transaction_id($_POST[EKASHU_TRANSACTION_ID]);
        }
        $order->payment_complete();
        $order->add_order_note( __( 'Order payment received', MOREFLO_CHECKOUT ) );//New translation
        $this->write_Woo_logs('Payment completed');//Logs
        if($event_type == 'success_payment_nmi'){
            $this->write_Woo_logs($order_id.'- successful payment redirect');//Logs
            WC()->cart->empty_cart();
            $thankyouURL = $order->get_checkout_order_received_url();
            wp_safe_redirect($thankyouURL);
        }

        if($event_type == SUCCESS){//Callback success method
            $this->write_Woo_logs($order_id.'- successful callback response recived');//Logs
            $returnResult[SUCCESS] = true;
            $returnResult['successCode'] = "200";

            wp_send_json( $returnResult, $returnResult['successCode'] );//Send received notification to server
        }

    }

    public function subcriptions_order($order_id)
    {
        $order = new WC_Order($order_id);

        if(function_exists('wcs_order_contains_subscription')){
            $subscriptionValidate = wcs_order_contains_subscription( $order );
            if($subscriptionValidate === true)
            {
                $this->write_Woo_logs('Order contain subcriptions products');
                $subs_order_id;
                $subscriptions = wcs_get_subscriptions_for_order($order_id, array( ORDER_TYPE => 'any' ));
                foreach( $subscriptions as $subscription_obj )
                {
                    $subs_order_id = $subscription_obj->get_id(); //current subscribtion id
                }

                $hash_key = $this->hash_key;
                $card_hash = $_POST['ekashu_card_hash'];
                $card_ref = $_POST['ekashu_card_reference'];
                $computedcardHash = base64_encode($subs_order_id.$card_hash);
                $computedcardrefHash = base64_encode($hash_key.$card_ref);

                update_post_meta($order_id, 'ekashu_enc_hash', $computedcardHash);//Adding transaction id
                update_post_meta($order_id, 'ekashu_enc_reference', $computedcardrefHash);//Adding transaction id

                $this->write_Woo_logs(sprintf('Order contain subcriptions products order Id# %s, subscription Id# %s', $order_id, $subs_order_id));

                unset($computedcardHash);
                unset($computedcardrefHash);
            }
        }
    }


    function validate_order_has_trial( $order ) {

        if ( ! $order instanceof WC_Order ){
            return new WP_Error( 'invalid_order', 'This order is invalid.' );
        }

        // Grab all items in an order
        if ( $items = $order->get_items() ) {

            // Loop through the items
            foreach( $items as $item_values ) {
                $product_id =  $item_values->get_product_id();

                $trial_length = get_post_meta( $product_id, '_subscription_trial_length', true );
                if($trial_length > 0)
                {
                    return true;
                }

            }
            // Otherwise, we return false
            return false;
        }
    }

    public function scheduled_NMI_subscription_payment($amount_to_charge, $renewal_order )
    {
        $reponse = $this->process_NMI_subscription_payment( $renewal_order, $amount_to_charge );//Process subcriptionS
        if((string)$reponse->Result->LocalResult == 0)
        {
            $renewal_order->add_order_note( __( 'Order renewal payment received', MOREFLO_CHECKOUT ) );//New translation

            $paymentTrId = (string)$reponse->TransactionDetails->CardEaseReference;
            if(isset($paymentTrId)){
                $renewal_order->payment_complete($paymentTrId);
                update_post_meta($renewal_order->get_id(), EKASHU_TRANSACTION_ID, $paymentTrId);//Adding transaction id
            }else{
                $renewal_order->payment_complete();
            }

            $this->write_Woo_logs('Order renewal completed');
        }
        if((string)$reponse->Result->LocalResult == 1)
        {
            $renewal_order->add_order_note( sprintf( __( 'Order renewal is cancelled due to %s reason', MOREFLO_CHECKOUT), (string)$reponse->Result->Errors->Error) );//New translation
            $renewal_order->update_status( 'wc-cancelled');
            $this->write_Woo_logs(sprintf('Order renewal is cancelled and response from xml protocol resultcode %s and error# %s.', (string)$reponse->Result->LocalResult, (string)$reponse->Result->Errors->Error));
        }
        if((string)$reponse->Result->LocalResult == 2)
        {
            $renewal_order->add_order_note( sprintf( __( 'Order renewal pending due to %s reason', MOREFLO_CHECKOUT), (string)$reponse->Result->Errors->Error) );//New translation
            $renewal_order->update_status( 'wc-pending');
            $this->write_Woo_logs(sprintf('Order renewal pending and response from xml protocol resultcode %s and error# %s.', (string)$reponse->Result->LocalResult, (string)$reponse->Result->Errors->Error));
        }
        if((string)$reponse->Result->LocalResult == 9)
        {
            $renewal_order->add_order_note( sprintf( __( 'Sorry, we are unable to process your payment at this time. Please reactivate and Please retry later. Order renewal update as pending due to %s reason', MOREFLO_CHECKOUT), (string)$reponse->Result->Errors->Error) );//New translation
            $renewal_order->update_status( 'wc-pending');
            $this->write_Woo_logs(sprintf('Order renewal pending and response from xml protocol resultcode %s and error# %s.', (string)$reponse->Result->LocalResult, (string)$reponse->Result->Errors->Error));
        }
    }

    public function retrieveParentSubcriptionsID($order_id)
    {
        $parent_subs_id;
        $subscriptions = wcs_get_subscriptions_for_order($order_id, array( ORDER_TYPE => 'any' ));
        foreach( $subscriptions as $subscription_obj )
        {
            $parent_subs_id = $subscription_obj->get_parent_id(); // This is subscription parent id
            if(!empty($parent_subs_id))
            {
                return $parent_subs_id;
            }
        }
    }

    public function process_NMI_subscription_payment($order, $amount_to_charge)
    {
        $hash_key = $this->hash_key;
        $order_id = $order->get_id();
        $order_currency = $order->get_currency();
        $subscriptions = wcs_get_subscriptions_for_order($order_id, array( ORDER_TYPE => 'any' ));
        $parent_id;
        $subcriptionid;

        $this->write_Woo_logs(sprintf('Start processing NMI subscription. Date time %s order-id %s',date(DATETIME_FORMATE), $order_id));

        foreach( $subscriptions as $subscription_obj )
        {
            $subcriptionid = $subscription_obj->get_id(); //current subscribtion id
            $parent_id = $subscription_obj->get_parent_id(); //parent id if needed
        }

        if(!empty($parent_id))
        {
            $cardHash = get_post_meta($parent_id, 'ekashu_enc_hash', true);
            $cardReference = get_post_meta($parent_id, 'ekashu_enc_reference', true);
        }

        $ekashuCardHash = str_replace($subcriptionid,"",base64_decode($cardHash));
        $ekashuCardRef = str_replace($hash_key,"",base64_decode($cardReference));

        $this->write_Woo_logs(sprintf('Processing order renewal payment request for order-id %s amount %s order parent-id %s subscription id %s ',$order_id,$amount_to_charge,$parent_id,$subcriptionid));

        $xmlObject = $this->prepareXMLObject($ekashuCardHash, $ekashuCardRef, $amount_to_charge, $order_currency);
        // $this->write_Woo_logs($xmlObject);
        $response  = $this->StartTransaction($xmlObject);
        libxml_use_internal_errors(true);
        $XMLtoObject = simplexml_load_string($response);
        if (!$XMLtoObject)
        {
            libxml_clear_errors();
            if ( $this->retry_interval <= 3 ) {
                $this->write_Woo_logs(sprintf('Retry NMI subscription date-time %s order-id %s',date(DATETIME_FORMATE), $order_id));
                sleep( $this->retry_interval );
                $this->retry_interval++;
                return $this->process_NMI_subscription_payment($order, $amount_to_charge);
            }
            $XMLtoObject = new stdClass();
            $XMLtoObject->Result->LocalResult = 9;
            $XMLtoObject->Result->Errors->Error = $response;

        }
        $this->write_Woo_logs(sprintf('Completed processing NMI subscription. Date-time %s order-id %s',date(DATETIME_FORMATE), $order_id));
        return $XMLtoObject;
    }

    public function StartTransaction($xmlObject)
    {
        if($this->sandbox == "yes"){
            $url = "https://test.cardeasexml.com/generic.cex";
        }else{
            $url = "https://live.cardeasexml.com/generic.cex";
        }

        $headers = array(
            "Content-type: text/xml",
            "Content-length: " . strlen($xmlObject),
            "Connection: close",
        );

        $ch = curl_init();
        curl_setopt($ch, CURLOPT_URL,$url);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($ch, CURLOPT_TIMEOUT, 10);
        curl_setopt($ch, CURLOPT_POST, true);
        curl_setopt($ch, CURLOPT_POSTFIELDS, $xmlObject);
        curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);

        $data = curl_exec($ch);
        if(curl_errno($ch))
        {
            $response = curl_error($ch);
        }
        else
        {
            curl_close($ch);
            $response = $data;
        }

        return $response;
    }

    public function prepareXMLObject($ekashuCardHash, $ekashuCardRef, $amount_to_charge, $order_currency)
    {
        $ekashu_terminal_id = $this->terminal_id;
        $ekashu_transaction_key = $this->transaction_key;

        $amount_to_charge = $this->formateAmount($amount_to_charge);
        $woodecimaloptions = get_option('woocommerce_price_num_decimals');

        if($woodecimaloptions > 0)
        {
            $amount_to_charge = str_replace('.','',$amount_to_charge);
        }

        return '<?xml version="1.0" encoding="us-ascii" standalone="yes"?>
                <Request type="CardEaseXML" version="1.0.0">
                    <TransactionDetails>
                        <MessageType autoconfirm="true">Auth</MessageType>
                        <Amount>'.$amount_to_charge.'</Amount>
                        <CurrencyCode>'.$order_currency.'</CurrencyCode>
                    </TransactionDetails>
                    <TerminalDetails>
                        <TerminalID>'.$ekashu_terminal_id.'</TerminalID>
                        <TransactionKey>'.$ekashu_transaction_key.'</TransactionKey>
                        <Software version="1.0.0">CardEaseXMLClient</Software>
                    </TerminalDetails>
                    <CardDetails>
                        <Manual type="cnp">
                            <CardReference>'.$ekashuCardRef.'</CardReference>
                            <CardHash>'.$ekashuCardHash.'</CardHash>
                        </Manual>
                    </CardDetails>
                </Request>';
    }

	/**
	 * Process refund.
	 *
	 * If the gateway declares 'refunds' support, this will allow it to refund.
	 * a passed in amount.
	 *
	 * @param  int    $order_id Order ID.
	 * @param  float  $amount Refund amount.
	 * @param  string $reason Refund reason.
	 * @return boolean True or false based on success, or a WP_Error object.
	 */
	public function process_refund( $order_id, $amount_to_refund = null, $reason = '' ) {
        $url = ( "yes" == $this->sandbox )
            ? 'https://test.cardeasexml.com/generic.cex'
            : 'https://live.cardeasexml.com/generic.cex';

        if ( ! class_exists( 'SimpleXMLElement' ) ) {
            return new WP_Error( 'moreflo-checkout-nmi-pay', sprintf( __( '%s is not installed on this server. Please contact your server administrator.', 'moreflo-checkout' ), 'SimpleXMLElement' ) .' '. __( 'Please log in to NMI to perform refunds.', 'moreflo-checkout' ) );
        }
        if ( ! function_exists( 'curl_init' ) ) {
            return new WP_Error( 'moreflo-checkout-nmi-pay', sprintf( __( '%s is not installed on this server. Please contact your server administrator.', 'moreflo-checkout' ), 'Curl' ) .' '. __( 'Please log in to NMI to perform refunds.', 'moreflo-checkout' ) );
        }

        $ekashu_transaction_id = get_post_meta( $order_id, 'ekashu_transaction_id', true );
        if ( ! $ekashu_transaction_id ) {
            return new WP_Error( 'moreflo-checkout-nmi-pay', __( 'Ekashu transaction ID is missing.', 'moreflo-checkout' ) .' '. __( 'Please log in to NMI to perform refunds.', 'moreflo-checkout' ) );
        }

        if ( is_null( $amount_to_refund ) ) {
            return new WP_Error( 'moreflo-checkout-nmi-pay', __( 'Amount to refund missing.', 'moreflo-checkout' ) .' '. __( 'Please log in to NMI to perform refunds.', 'moreflo-checkout' ) );
        }

        $order = wc_get_order( $order_id );
        $order_currency = $order->get_currency();

        $amount_to_refund = $this->formateAmount( $amount_to_refund );
        $woodecimaloptions = get_option( 'woocommerce_price_num_decimals' );

        if ( $woodecimaloptions > 0 ) {
            $amount_to_refund = str_replace( '.', '', $amount_to_refund );
        }

        $requestXml = sprintf(
            "<?xml version='1.0' encoding='UTF-8' standalone='yes'?>
            <Request type='CardEaseXML' version='1.2.0'>
                <TransactionDetails>
                    <MessageType>Refund</MessageType>
                    <CardEaseReference>%s</CardEaseReference>
                    %s
                </TransactionDetails>
                <TerminalDetails>
                    <TerminalID>%s</TerminalID>
                    <TransactionKey>%s</TransactionKey>
                    <Software version='1.3.6'>MoreFloCheckout</Software>
                    <Component version='1.0.0'>NmiPayGateway</Component>
                </TerminalDetails>
            </Request>",
            $ekashu_transaction_id,
            "<Amount>{$amount_to_refund}</Amount><CurrencyCode>{$order_currency}</CurrencyCode>",
            $this->terminal_id,
            $this->transaction_key
        );

        $ch = curl_init();
	    curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, 100);
	    curl_setopt($ch, CURLOPT_HTTPHEADER, array('Content-Type: text/xml'));
	    curl_setopt($ch, CURLOPT_POST, TRUE);
	    curl_setopt($ch, CURLOPT_RETURNTRANSFER, TRUE);
	    curl_setopt($ch, CURLOPT_TIMEOUT, 60);
	    curl_setopt($ch, CURLOPT_URL, $url);
	    curl_setopt($ch, CURLOPT_USERAGENT, 'MoreFloCheckout NmiPayGateway');
        curl_setopt($ch, CURLOPT_POSTFIELDS, $requestXml);

        $response = curl_exec($ch);
        if ( curl_errno( $ch ) ) {
            $this->write_Woo_logs( 'Error refunding order '. $order_id .'. Request: '. var_export( $requestXml, true ) );
            $this->write_Woo_logs( 'Error refunding order '. $order_id .'. Curl Error: '. var_export( curl_error( $ch ), true ) );
            return new WP_Error( 'moreflo-checkout-nmi-pay', __( 'Unknown error while refunding.', 'moreflo-checkout' ) .' '. __( 'Please log in to NMI to perform refunds.', 'moreflo-checkout' ) );
        }

        curl_close($ch);

        // Parse response
        $xml = new SimpleXMLElement( $response );
        $result = (string) $xml->Result->LocalResult;
        if ( '0' != $result ) {
            // Errors from NMI, abort
            $errors = [];
            foreach ( $xml->Result->Errors->Error as $error ) {
                $errors[] = sprintf( __( 'Error %s: %s', 'moreflo-checkout' ), (string) $error['code'], (string) $error );
            }

            return new WP_Error( 'moreflo-checkout-nmi-pay', implode( '; ', $errors ) );
        }

        // Refund success, save reference
        add_post_meta( $order_id, 'mf_checkout_nmi_refunds', $response );

        return true;
	}

    public function write_Woo_logs($details)
    {
        if ( $this->log_enabled == 'yes' ) {//record if only log enable this method to run
            $this->logger->debug(wc_print_r("************", true), $this->logger_context);
            $this->logger->debug(wc_print_r($details, true), $this->logger_context);
            $this->logger->debug(wc_print_r("************", true), $this->logger_context);
        }
    }

    public function formateAmount($amount)
    {
        // Adding decimal point zeros if required based on woocommerece settings
        $woodecimaloptions = get_option('woocommerce_price_num_decimals');

        $nmiamountformats = [0 => '00', 1 => '0'];
        foreach ($nmiamountformats as $key => $value) {
            if($key == $woodecimaloptions)
            {
                $amount = $amount.$value;
            }
        }
        if($woodecimaloptions > 0)
        {
            if($woodecimaloptions >= 3)
            {
                $amount = substr($amount, 0, -($woodecimaloptions - 2) );
            }
        }

        return $amount;
    }
}

class NmiPayHashInput {
    public static function validate_hash( $data, $endpoint ) {
        $post_data = null;

        if ( ! isset( $data['ekashu_hash_code'] ) ) { return false; }
        if ( ! isset( $data['ekashu_hash_code_format'] ) ) { return false; }
        if ( ! isset( $data['ekashu_hash_code_result'] ) ) { return false; }
        if ( ! isset( $data['ekashu_hash_code_result_format'] ) ) { return false; }
        if ( ! isset( $data['ekashu_hash_code_result_type'] ) ) { return false; }
        if ( ! isset( $data['ekashu_hash_code_result_version'] ) ) { return false; }
        if ( ! isset( $data['ekashu_hash_code_type'] ) ) { return false; }
        if ( ! isset( $data['ekashu_hash_code_version'] ) ) { return false; }

        foreach ( $data as $name => $value ) {
            if ( preg_match( '/ekashu_/', $name ) == 1 ) {
                $post_data .= urlencode( $name ) .'='. urlencode( $value );
                $post_data .= '&';
            }
        }

        $post_data = substr( $post_data, 0, -1 );

        $response = wp_remote_post( $endpoint .'/validate_hash_code.php', [
            'body' => $post_data,
        ] );

        return 200 == wp_remote_retrieve_response_code( $response );
    }

    public function __construct( $data = [] ) {
        $required_keys = [ 'ekashu_amount', 'ekashu_currency', 'ekashu_reference', 'ekashu_seller_id', 'ekashu_seller_key' ];
        if ( empty( $data ) || ! $this->array_has_all_keys( $required_keys, $data ) ) {
            throw new Exception('Bad Nmi hash generation', 1);
        }

        foreach ( $data as $key => $value ) {
            $this->{$key} = $value;
        }
    }

    public function get_hash( $hash_key ) {
        $hash_code_input = (array)$this;
        ksort( $hash_code_input );

        $hash_code = null;
        foreach ( $hash_code_input as $name => $value ) {
            $hash_code .= trim( $value );
            $hash_code .= '&';
        }

        $hash_code = substr( $hash_code, 0, -1 );

        return base64_encode( hash_hmac( 'sha256', $hash_code, $hash_key, true ) );
    }

    public function get_form_data() {
        $hash_code_input = (array)$this;
        ksort( $hash_code_input );

        return $hash_code_input;
    }

    public $ekashu_3d_secure_verify = null;
    public $ekashu_amount = null;
    public $ekashu_amount_format = null;
    public $ekashu_auto_confirm = null;
    public $ekashu_callback_failure_url = null;
    public $ekashu_callback_include_post = null;
    public $ekashu_callback_success_url = null;
    public $ekashu_card_address_editable = null;
    public $ekashu_card_address_required = null;
    public $ekashu_card_address_verify = null;
    public $ekashu_card_email_address_mandatory = null;
    public $ekashu_card_phone_number_mandatory = null;
    public $ekashu_card_title_mandatory = null;
    public $ekashu_card_zip_code_verify = null;
    public $ekashu_currency = null;
    public $ekashu_delivery_address_editable = null;
    public $ekashu_delivery_address_required = null;
    public $ekashu_delivery_email_address_mandatory = null;
    public $ekashu_delivery_phone_number_mandatory = null;
    public $ekashu_delivery_title_mandatory = null;
    public $ekashu_description = null;
    public $ekashu_device = null;
    public $ekashu_duplicate_check = null;
    public $ekashu_duplicate_minutes = null;
    public $ekashu_failure_return_text = null;
    public $ekashu_failure_url = null;
    public $ekashu_hash_code_format = null;
    public $ekashu_hash_code_type = 'SHA256HMAC';
    public $ekashu_hash_code_version = '2.0.0';
    public $ekashu_include_post = null;
    public $ekashu_invoice_address_editable = null;
    public $ekashu_invoice_address_required = null;
    public $ekashu_invoice_email_address_mandatory = null;
    public $ekashu_invoice_phone_number_mandatory = null;
    public $ekashu_invoice_title_mandatory = null;
    public $ekashu_locale = null;
    public $ekashu_payment_methods = null;
    public $ekashu_reference = null;
    public $ekashu_request_type = null;
    public $ekashu_return_text = null;
    public $ekashu_seller_address = null;
    public $ekashu_seller_email_address = null;
    public $ekashu_seller_id = null;
    public $ekashu_seller_key = null;
    public $ekashu_seller_name = null;
    public $ekashu_shortcut_icon = null;
    public $ekashu_title = null;
    public $ekashu_style_sheet = null;
    public $ekashu_success_url = null;
    public $ekashu_verification_value_mask = null;
    public $ekashu_verification_value_verify = null;
    public $ekashu_viewport = null;

    private function array_has_all_keys( $keys, $arr ) {
        if ( empty( $keys ) ) {
            return true;
        }

        foreach ( $keys as $key ) {
            if ( ! isset( $arr[ $key ] ) ) {
                return false;
            }
        }

        return true;
    }
}
