<?php

if ( ! defined( 'ABSPATH' ) ) {
	exit; // Exit if accessed directly
}

/**
 * Credit Card Payments for Woocommerce with CyberSource implementation class.
 *
 * @class 		Invoice_Pay
 * @extends		WC_Payment_Gateway
 * @version		1.0.0
 */

require_once plugin_dir_path(__FILE__) . 'classes/api/class-woocommerce-swish-pay-client.php';
require_once plugin_dir_path(__FILE__) . 'classes/api/class-woocommerce-swish-pay-connection.php';
require_once plugin_dir_path(__FILE__) . 'classes/api/class-woocommerce-swish-pay-error.php';

class Swish_Pay_Gateway extends WC_Payment_Gateway {

	/**
	 * Current Swish environment.
	 *
	 * @var string $swish_environment
	 */
    private $swish_environment;

	/**
	 * Path to root CA file.
	 *
	 * @var string $root_certificate
	 */
    private $root_certificate;

	/**
	 * Path to public certificate file (.PEM or .p12).
	 *
	 * @var string $public_certificate
	 */
    private $public_certificate;

	/**
	 * Path to private certificate file (.key).
	 *
	 * @var string $private_certificate
	 */
    private $private_certificate;

	/**
	 * Password for certificate.
	 *
	 * @var string $private_key
	 */
    private $private_key;

	/**
	 * Whether logging test is enabled, 'yes' or 'no'.
	 *
	 * @var string $log_enabled
	 */
    private $log_enabled;

	/**
	 * Whether connection test is enabled, 'yes' or 'no'.
	 *
	 * @var string $swish_connection_test_enabled
	 */
    private $swish_connection_test_enabled;

	/**
	 * Phone number of payer used for test connection.
	 *
	 * @var string $swish_connection_test_payer
	 */
    private $swish_connection_test_payer;

	/**
	 * Phone number of payer.
	 *
	 * @var string $payer_number
	 */
    private $payer_number;

	/**
	 * API URL for current Swish Environment.
	 *
	 * @var string $api_end_URL
	 */
    private $api_end_URL;

	/**
	 * Phone number of payer.
	 *
	 * @var string $payee_number
	 */
    private $payee_number;

	/**
	 * @var WC_Logger $logger
	 */
	private $logger;

	/**
	 * @var array $logger
	 */
	private $logger_context;

	/**
	 * @var array $handled_files
	 */
	private $handled_files;

    /**
     * Class constructor, more about it in Step 3
     */
    public function __construct() {
        $this->id                 = 'swish_pay_gateway';
		$this->method_title       = __( 'MoreFlo Checkout', 'moreflo-checkout' );
        $this->method_description = __( 'Use Swish payment as payment gateway. Requires SSL and generate certificate from swish.', 'moreflo-checkout' );
        $this->has_fields         = true;
        $this->supports           = array(
            'products',
            'refunds',
        );

        $this->logger = wc_get_logger();
        $this->logger_context = array( 'source' => $this->id );

		$this->check_user_cert_upload_dir();
	    $this->handle_gateway_update();

        // Get setting values
        $this->title             = $this->get_option( 'title' );
        $this->swish_environment = $this->get_option( 'swish_environment' );

		switch ( $this->swish_environment ) {
			case 'mss':
				$this->api_end_URL         = 'mss.cpc.getswish.net';
				$this->root_certificate    = DIRPATH . 'certs/swish/CA/RootCA-prod_mss.pem';
				$this->public_certificate  = DIRPATH . 'certs/swish/MSS/bundle_mss_1.9.3.p12';
				$this->private_certificate = '';
				$this->private_key         = 'swish';
				$this->payee_number        = '1234679304';
				break;
			case 'sandbox':
				$this->api_end_URL         = 'staging.getswish.pub.tds.tieto.com';
				$this->root_certificate    = DIRPATH . 'certs/swish/CA/RootCA-sandbox.pem';
				$this->public_certificate  = $this->get_option( self::get_cert_info_by_type( 'sandbox_public', 'option_name' ) );
				$this->private_certificate = $this->get_option( self::get_cert_info_by_type( 'sandbox_private', 'option_name' ) );
				$this->private_key         = $this->get_option( 'sandbox_private_key' );
				$this->payee_number        = $this->get_option( 'sandbox_payee_number' );
				break;
			case 'prod':
			default:
				$this->api_end_URL      = 'cpc.getswish.net';
				$this->root_certificate = DIRPATH . 'certs/swish/CA/RootCA-prod_mss.pem';
                $this->payee_number     = $this->get_option( 'payee_number' );

                if (
                    'yes' === $this->get_option( 'swish_user_server_certs' ) &&
                    self::nmf_certs_exist()
                ) {
	                $this->public_certificate  = NMF_CHECKOUT_SWISH_PEM_PATH;
	                $this->private_certificate = NMF_CHECKOUT_SWISH_KEY_PATH;
	                $this->private_key         = defined( 'NMF_CHECKOUT_SWISH_KEY_PASS' ) ? NMF_CHECKOUT_SWISH_KEY_PASS : false;
                } else {
                    $this->public_certificate  = $this->get_option( self::get_cert_info_by_type( 'public', 'option_name' ) );
                    $this->private_certificate = $this->get_option( self::get_cert_info_by_type( 'private', 'option_name' ) );
                    $this->private_key         = $this->get_option( 'live_private_key' );
                }


				break;

		}

        $this->log_enabled = $this->get_option( 'log_enabled' );

	    $this->swish_connection_test_enabled = $this->get_option( 'swish_connection_test_enabled' );
	    $this->swish_connection_test_payer   = $this->get_option( 'test_payer_number' );

        $this->init_form_fields();

        //Save settings fields
        add_action( 'woocommerce_update_options_payment_gateways_' . $this->id, array( $this, 'process_admin_options' ), 10, 0);
        add_action( 'woocommerce_update_option', array( $this, 'process_admin_file_options' ));

		add_action( 'admin_enqueue_scripts', array( $this, 'admin_scripts' ) );

		add_action( 'woocommerce_settings_checkout', array( $this, 'display_swish_errors' ) );

		add_action('woocommerce_api_swish_pay_gateway_connection_test', array($this, 'swish_connection_test_callback_response'), 10, 0);

		if ( 'yes' === $this->enabled ) {
            add_action('woocommerce_api_swish_pay_gateway_wait', array($this, 'swish_wait_callback_response'), 10, 0);
            add_action('woocommerce_api_swish_pay_gateway', array($this, 'swish_callback_response'), 10, 0);
            add_action('woocommerce_api_swish_pay_gateway', array($this, 'swish_callback_response_local_dev'), 9, 0);

            add_action( 'wp_enqueue_scripts', array($this, 'thankyou_scripts'), 10 );
            add_action( 'woocommerce_before_thankyou', array($this, 'moreflo_thankyou'),10, 1);
        }
    }

	private function check_user_cert_upload_dir() {
		$dir        = self::get_user_certs_path();
		$upload_dir = wp_mkdir_p( $dir ) ? $dir : false;

		if ( ! $upload_dir ) {
			return false;
		}

		if ( ! file_exists( $dir . '/.htaccess' ) ) {
			return copy( DIRPATH . 'certs/.htaccess', $dir . '/.htaccess' );
		}

		return true;
	}

	public static function get_user_certs_path( $cert_type = null ) {
		$wp_uploads = wp_upload_dir( null, false );
		return $wp_uploads['basedir'] . '/moreflo-checkout/swish-certs';
	}

	private static function get_cert_info_by_type( $type, $key = null ) {
		$info = array();

		$cert_uploads_path = self::get_user_certs_path();

		switch ( $type ) {
			case 'public':
				$info['option_name'] = 'live_certificate';
				$info['filename']    = 'prod_cert_file';
				$info['filepath']    = $cert_uploads_path . '/' . $info['filename'];
				break;

			case 'private':
				$info['option_name'] = 'live_private_certificate';
				$info['filename']    = 'prod_priv_cert_file';
				$info['filepath']    = $cert_uploads_path . '/' . $info['filename'];
				break;

			case 'sandbox_public':
				$info['option_name'] = 'sandbox_certificate';
				$info['filename']    = 'sandbox_cert_file';
				$info['filepath']    = $cert_uploads_path . '/' . $info['filename'];
				break;

			case 'sandbox_private':
				$info['option_name'] = 'sandbox_private_certificate';
				$info['filename']    = 'sandbox_priv_cert_file';
				$info['filepath']    = $cert_uploads_path . '/' . $info['filename'];
				break;

			case 'csr':
				$info['option_name'] = '';
				$info['filename']    = 'prod_swish.csr';
				$info['filepath']    = $cert_uploads_path . '/' . $info['filename'];
				break;
		}

		if ( ! $key ) {
			return $info;
		}

		return ! empty( $info[ $key ] ) ? $info[ $key ] : '';
	}


	/**
	 * Handles update for Swish gateway settings.
	 *
	 * @return void
	 */
	public function handle_gateway_update() {

		if ( MOREFLO_CHECKOUT_VERSION === get_option( 'mf_checkout_version' ) ) {
			return; // Already updated.
		}

		/**
		 * Backwards compatibility for plugin pre v1.7.0.
		 */
		if ( version_compare( get_option( 'mf_checkout_version' ), '1.7.0', '<' ) ) {
			/**
			 * Check Swish Environment.
			 */
			$old_sandbox_val = $this->get_option( 'sandbox' );
			$old_mss_val     = $this->get_option( 'sandbox_mss' );

			if ( 'no' !== $old_mss_val ) {
				$old_env = 'mss';
			} elseif ( 'no' !== $old_sandbox_val ) {
				$old_env = 'sandbox';
			} else {
				$old_env = 'prod';
			}

			$this->update_option( 'swish_environment', $old_env );

			/**
			 * Check Swish payee number.
			 */
			if ( empty( $this->get_option( 'sandbox_payee_number' ) ) ) {
				$this->update_option( 'sandbox_payee_number', $this->get_option( 'payee_number' ) );
			}


			/**
			 * Check certificates.
			 */
			// Check if there is a .key file with the same name as the public certificate file for production settings.
			$prod_public_cert = $this->get_option( 'live_certificate' );
			if ( $prod_public_cert ) {
				$prod_public_cert_info = pathinfo( $prod_public_cert );
				$prod_private_cert_path = $prod_public_cert_info['dirname'] . '/' . $prod_public_cert_info['filename'] . '.key';

				if ( file_exists( $prod_private_cert_path ) ) {
					$this->update_option( 'live_private_certificate', $prod_private_cert_path );
				}
			}

			// Check if there is a .key file with the same name as the public certificate file for sandbox settings.
			$sandbox_public_cert = $this->get_option( 'sandbox_certificate' );
			if ( $sandbox_public_cert ) {
				$sandbox_public_cert_info = pathinfo( $sandbox_public_cert );
				$sandbox_private_cert_path = $sandbox_public_cert_info['dirname'] . '/' . $sandbox_public_cert_info['filename'] . '.key';

				if ( file_exists( $sandbox_private_cert_path ) ) {
					$this->update_option( 'sandbox_private_certificate', $sandbox_private_cert_path );
				}

			}
		}

		/**
		 * Backwards compatibility for plugin pre v1.7.2.
		 */
		if ( version_compare( get_option( 'mf_checkout_version' ), '1.7.2', '<' ) ) {
			// Copy the old certificates to the new location.
			$this->copy_cert_file( 'public' );
			$this->copy_cert_file( 'private' );
			$this->copy_cert_file( 'sandbox_public' );
			$this->copy_cert_file( 'sandbox_private' );
		}

		/**
		 * Update finished.
		 */
		update_option( 'mf_checkout_version', MOREFLO_CHECKOUT_VERSION );
	}

	/**
	 * @param string $cert_type Type of certificate file to copy.
	 *
	 * @return bool Whether the file was copied or not.
	 */
	private function copy_cert_file( $cert_type, $copy_from = null, $copy_to = null ) {
		$cert_info = self::get_cert_info_by_type( $cert_type );

		if ( $copy_from ) {
			$current_path = $copy_from;
		} else {
			if ( empty( $cert_info['option_name'] ) ) {
				return false;
			}
			$current_path = $this->get_option( $cert_info['option_name'] );
		}

		if ( ! is_file( $current_path ) ) {
			return false;
		}

		if ( $copy_to ) {
			$new_path = $copy_to;
		} else {
			if ( empty( $cert_info['filepath'] ) ) {
				return false;
			}
			$path_parts = pathinfo( $current_path );
			$new_path   = $cert_info['filepath'] . '.' . $path_parts['extension'];
		}

		if ( $new_path !== $current_path ) {
			if ( ! copy( $current_path, $new_path ) ) {
				return false;
			}

			! empty( $cert_info['option_name'] ) && $this->update_option( $cert_info['option_name'], $new_path );
			return true;
		}
		return false;
	}

    public static function nmf_certs_exist(): bool{
        if (
            defined( 'NMF_CHECKOUT_SWISH_PEM_PATH' ) &&
            defined( 'NMF_CHECKOUT_SWISH_KEY_PATH' ) &&
            file_exists( NMF_CHECKOUT_SWISH_PEM_PATH ) &&
            file_exists( NMF_CHECKOUT_SWISH_KEY_PATH )
        ) {
            return true;
        }
        return false;
    }

	/**
	 * Plugin options, we deal with it in Step 3 too
	 */
	public function init_form_fields(){
	    $log_enabled_description = '';

	    if ( defined( 'WC_LOG_DIR' ) ) {
			$swish_log_url = add_query_arg(
				array(
					'page'     => 'wc-status',
					'tab'      => 'logs',
					'log_file' => WC_Log_Handler_File::get_log_file_name( $this->id ),
				),
				admin_url( 'admin.php' )
			);

			// Translators: %1$s and %2$s are placeholders for HTML tags.
	        $log_enabled_description = sprintf( __( '%1$sView Log%2$s', 'moreflo-checkout' ), '<a href="' . esc_url( $swish_log_url ) . '" target="_blank">', '</a>' );
	    }

		$gateway_environment_desc =
			// translators: %1$s and %2$s are placeholders for the opening and closing <strong> tags, respectively.
			sprintf( __( '%1$sProduction (Live): %2$s Use this mode for live transactions.', 'moreflo-checkout' ), '<strong>', '</strong>' ) . '<br>' .
			// translators: %1$s and %2$s are placeholders for the opening and closing <strong> tags, respectively.
			sprintf( __( '%1$sMerchant Swish Simulator: %2$s For basic testing.', 'moreflo-checkout' ), '<strong>', '</strong>' ) . '<br>' .
			// translators: %1$s and %2$s are placeholders for the opening and closing <strong> tags, respectively.
			sprintf( __( '%1$sSwish Sandbox: %2$s For advanced verification of the full payment request flow. A Sandbox account from Swish is required.', 'moreflo-checkout' ), '<strong>', '</strong>' ) . '<br><br>' .
			// translators: %1$s and %2$s are placeholders for the opening and closing <a> tags, respectively.
			sprintf( __( 'Read more about Swish Environments %1$shere%2$s.', 'moreflo-checkout' ), '<a href="https://developer.swish.nu/documentation/environments" target="_blank">', '</a>' );

		$prod_public_file_desc =
			__( 'Upload .pem file downloaded from the Swish Portal (.p12 file can also be used).', 'moreflo-checkout' ) .
			$this->generate_cert_download_link( 'public' );

		$prod_private_file_desc =
			__( 'Upload your private .key certificate file. This key file is generated and automatically stored when generating a new certificate request above.', 'moreflo-checkout' ) .
			$this->generate_cert_download_link( 'private' );

		$connection_test_status = 'yes' === $this->swish_connection_test_enabled ? '<i class="spinner is-active" style="float: none;"></i>' : __( 'Disabled', 'moreflo-checkout' );

		$support_url = 'https://supportflo.northmill.com/s/article/Flo-Checkout-Ta-betalt-med-kort-i-Wordpress';
		$support_info = __( 'Read more on how to set this up on our support page:', 'moreflo-checkout' ) . '<br>' .
		                    sprintf('<a href="%s" target="_blank">%s</a>', $support_url, $support_url );


        $form_fields = array();

        $form_fields['enabled'] = array(
            'title'       => __( 'Enable/Disable', 'moreflo-checkout' ),
            'label'       => __( 'Enable Swish Payments', 'moreflo-checkout' ),
            'type'        => 'checkbox',
            'description' => '',
            'default'     => 'no'
        );

        $form_fields['title'] = array(
				'title'       => __( 'Title', 'moreflo-checkout' ),
				'type'        => 'text',
				'description' => __( 'Title for payment type displayed in checkout.', 'moreflo-checkout' ),
				'default'     => __( 'Swish Pay', 'moreflo-checkout' )
			);

        $form_fields['swish_environment'] = array(
            'title'       => __( 'Swish Environment', 'moreflo-checkout' ),
            'type'        => 'select',
            'description' => $gateway_environment_desc,
            'default'     => 'production',
            'options'     => array(
                'prod'    => __( 'Production', 'moreflo-checkout' ),
                'mss'     => __( 'Merchant Swish Simulator', 'moreflo-checkout' ),
                'sandbox' => __( 'Swish Sandbox', 'moreflo-checkout' ),
            ),
        );
        /**
         * Production.
         */
        $form_fields['swish_prod_settings'] = array(
            'title'       => __( 'Production settings', 'moreflo-checkout' ),
            'type'        => 'title',
        );

        if ( self::nmf_certs_exist() ) {
	        $form_fields['swish_user_server_certs'] = array(
		        'title'       => __( 'Use Server Swish Certificates', 'moreflo-checkout' ),
		        'label'       => __( 'Uses certificates stored on server instead of custom certificates.', 'moreflo-checkout' ),
		        'type'        => 'checkbox',
		        'description' => '',
		        'default'     => 'no'
	        );
        }
        $form_fields['swish_prod_settings_generate_csr'] = array(
            'title'       => '',
            'type'        => 'title',
            'description' => $this->get_certificate_generation_html(),
        );
        $form_fields[ self::get_cert_info_by_type('public', 'filename' ) ] = array(
            'title'       => __( 'Public certificate file', 'moreflo-checkout' ),
            'type'        => 'file',
            'description' => $prod_public_file_desc,
            'custom_attributes' => array(
                'no_value' => true
            )
        );
        $form_fields[ self::get_cert_info_by_type('private', 'filename' ) ] = array(
            'title'       => __( 'Private certificate file', 'moreflo-checkout' ),
            'type'        => 'file',
            'description' => $prod_private_file_desc,
            'custom_attributes' => array(
                'no_value' => true
            )
        );
        $form_fields['live_private_key'] = array(
            'title'       => __( 'Private key', 'moreflo-checkout' ),
            'type'        => 'text',
            'description' => __( 'Password for certificate.', 'moreflo-checkout' ),
            'default'     => 'swish'
        );
        $form_fields['payee_number'] = array(
            'title'       => __( 'Swish Number', 'moreflo-checkout' ),
            'type'        => 'text',
            'description' => __( 'Swish number you want to get paid to.', 'moreflo-checkout' ),
            'default'     => ''
        );
        /**
         * MSS.
         */
        $form_fields['swish_mss_settings'] = array(
            'title'       => __( 'Merchant Swish Simulator settings', 'moreflo-checkout' ),
            'description' => __( 'No extra setup needed when in Merchant Swish Simulator environment.', 'moreflo-checkout' ),
            'type'        => 'title',
        );
        /**
         * Sandbox settings.
         */
        $form_fields['swish_sandbox_settings'] = array(
            'title'       => __( 'Sandbox settings', 'moreflo-checkout' ),
            'type'        => 'title',
        );
        $form_fields[ self::get_cert_info_by_type( 'sandbox_public', 'filename' ) ] = array(
            'title'       => __( 'Public certificate upload', 'moreflo-checkout' ),
            'type'        => 'file',
            'description' => __( 'Upload your .pem or .p12 certificate file.', 'moreflo-checkout' ) .
                             $this->generate_cert_download_link( 'sandbox_public' ),
            'custom_attributes' => array(
                'no_value' => true
            )
        );
        $form_fields[ self::get_cert_info_by_type( 'sandbox_private', 'filename' ) ] = array(
            'title'       => __( 'Private certificate upload', 'moreflo-checkout' ),
            'type'        => 'file',
            'description' => __( 'Upload your private .key certificate file.', 'moreflo-checkout' ) .
                             $this->generate_cert_download_link( 'sandbox_private' ),
            'custom_attributes' => array(
                'no_value' => true
            )
        );
        $form_fields['sandbox_private_key'] = array(
            'title'       => __( 'Private key', 'moreflo-checkout' ),
            'type'        => 'text',
            'description' => __( 'Password for private certificate.', 'moreflo-checkout' ),
            'default'     => 'swish'
        );
        $form_fields['sandbox_payee_number'] = array(
            'title'       => __( 'Merchant Swish Number', 'moreflo-checkout' ),
            'type'        => 'text',
            'description' => __( 'Swish number receiving payments.', 'moreflo-checkout' ),
            'default'     => ''
        );
        /**
         * Connection test.
         */
        $form_fields['swish_connection_test'] = array(
            'title'       => __( 'Connection test', 'moreflo-checkout' ),
            'description' => '<strong>' . __( 'Test status', 'moreflo-checkout' ) . ':</strong><br><span id="swish-connection-test-result">' . $connection_test_status . '</span>',
            'type'        => 'title',
        );
        $form_fields['swish_connection_test_enabled'] = array(
            'title'       => __( 'Connection test', 'moreflo-checkout' ),
            'label'       => __( 'Enable connection test', 'moreflo-checkout' ),
            'description' => __( 'Enable to run connection test when this page is loaded.', 'moreflo-checkout' ) . '<br>' .
                             __( 'Connection will be based on current Swish Environment settings.', 'moreflo-checkout' ),
            'type'        => 'checkbox',
            'default'     => 'no'
        );
        $form_fields['test_payer_number'] = array(
            'title'       => __( 'Test Payer Number', 'moreflo-checkout' ),
            'type'        => 'text',
            // translators: %1$s and %2$s are placeholders for the HTML tags <strong> and </strong> respectively.
            'description' => sprintf( __( 'Optional phone number used for connection test, on a successful connection test a payment request will be made to this number. %1$sNote that when in Production environment, this request will be real. Do not accept the request unless you want to perform a real transaction.%2$sThe request is canceled automatically after a few seconds.', 'moreflo-checkout' ), '<br><strong style="color: red;">', '</strong><br>' ),
            'default'     => ''
        );
        $form_fields['support_info'] = array(
            'title'       => __( 'Guide and support', 'moreflo-checkout' ),
            'description' => $support_info,
            'type'        => 'title',
        );
        $form_fields['log_enabled'] = array(
            'title'       => __( 'Debug Log', 'moreflo-checkout' ),
            'label'       => __( 'Enable Logging', 'moreflo-checkout' ),
            'description' => $log_enabled_description,
            'type'        => 'checkbox',
            'default'     => 'no'
        );

        $this->form_fields = $form_fields;
	}

	/**
	 * Generate html for certificate genration button.
	 *
	 * @return string HTML for button.
	 */
	private function get_certificate_generation_html() {
		ob_start();
		esc_html_e( 'Click this button to generate a new certificate request to upload to Swish. This will invalidate any existing certificates created here.', 'moreflo-checkout' );
		?>
		<br>
		<?php
		echo sprintf(
			__( 'The generated certificate request file should be uploaded to the %1$sSwish Business Portal%2$s.', 'moreflo-checkout' ),
			'<a href="https://portal.swish.nu/" target="_blank">',
			'</a>'
		);
		?>
		<br><br>
		<button id="nm-generate-new-cert-open">
			<?php esc_html_e( 'Generate new certificate request', 'moreflo-checkout' ); ?>
		</button>
		<br><br>
		<?php
		echo $this->generate_cert_download_link( 'csr', false );
		return ob_get_clean();
	}

	/**
	 * Generate html for certificate download link.
	 *
	 * @param string $type          Type of certificate to download.
	 * @param bool   $add_br_before Whether to add line break before link or not.
	 * @return string HTML for link.
	 */
	public function generate_cert_download_link( $type, $add_br_before = true ) {

		switch ( $type ) {
			case 'public':
				$cert = $this->get_option( self::get_cert_info_by_type( 'public', 'option_name' ) );
				$text = __( 'Download current public certificate', 'moreflo-checkout' );
				$url  = admin_url( 'admin-post.php?action=download_cert&type=public' );
				break;

			case 'private':
				$cert = $this->get_option( self::get_cert_info_by_type( 'private', 'option_name' ) );
				$text = __( 'Download current private certificate', 'moreflo-checkout' );
				$url  = admin_url( 'admin-post.php?action=download_cert&type=private' );
				break;

			case 'csr':
				$cert = self::get_cert_info_by_type( 'csr', 'filepath' );
				$text = __( 'Download previous .csr file', 'moreflo-checkout' );
				$url  = admin_url( 'admin-post.php?action=download_cert&type=csr' );
				break;

			case 'sandbox_public':
				$cert = $this->get_option( self::get_cert_info_by_type( 'sandbox_public', 'option_name' ) );
				$text = __( 'Download current public certificate', 'moreflo-checkout' );
				$url  = admin_url( 'admin-post.php?action=download_cert&type=sandbox_public' );
				break;

			case 'sandbox_private':
				$cert = $this->get_option( self::get_cert_info_by_type( 'sandbox_private', 'option_name' ) );
				$text = __( 'Download current private certificate', 'moreflo-checkout' );
				$url  = admin_url( 'admin-post.php?action=download_cert&type=sandbox_private' );
				break;

			default:
				return '';
		}

		if ( ! file_exists( $cert ) ) {
			return '';
		}

		$link = sprintf(
			'<a href="%s" data-current="%s" target="_blank">%s</a>',
			$url,
			$cert,
			esc_html( $text ),
		);

		return $add_br_before ? '<br>' . $link : $link;
	}

	/**
	 * Generate html for certificate upload form.
	 *
	 * @param string $key Field key
	 * @param array  $data Field data
	 * @return string HTML for form.
	 */
	public function generate_file_html( $key, $data ) {
		$field_key = $this->get_field_key( $key );
		$defaults  = array(
			'title'             => '',
			'disabled'          => false,
			'class'             => '',
			'css'               => '',
			'placeholder'       => '',
			'type'              => 'text',
			'desc_tip'          => false,
			'description'       => '',
			'custom_attributes' => array(),
		);

		$data = wp_parse_args( $data, $defaults );

		$value = empty( $data['custom_attributes']['no_value'] ) ? $this->get_option( $key ) : '';

		ob_start();
		?>
		<tr valign="top">
			<th scope="row" class="titledesc">
				<label for="<?php echo esc_attr( $field_key ); ?>"><?php echo wp_kses_post( $data['title'] ); ?> <?php echo $this->get_tooltip_html( $data ); // WPCS: XSS ok. ?></label>
			</th>
			<td class="forminp">
				<fieldset>
					<legend class="screen-reader-text"><span><?php echo wp_kses_post( $data['title'] ); ?></span></legend>
					<input class="input-file <?php echo esc_attr( $data['class'] ); ?>" type="<?php echo esc_attr( $data['type'] ); ?>" name="<?php echo esc_attr( $field_key ); ?>" id="<?php echo esc_attr( $field_key ); ?>" style="<?php echo esc_attr( $data['css'] ); ?>" value="<?php echo esc_attr( $value ); ?>" <?php disabled( $data['disabled'], true ); ?> <?php echo $this->get_custom_attribute_html( $data ); // WPCS: XSS ok. ?> />
					<?php echo $this->get_description_html( $data ); // WPCS: XSS ok. ?>
				</fieldset>
			</td>
		</tr>
		<?php

		return ob_get_clean();
	}

	/**
	 * Process settings file uploads.
	 *
	 * @param array $option_data
	 *
	 * @return void
	 */
	public function process_admin_file_options( $option_data ) {
		if (
			! empty( $option_data['id'] ) &&
			"woocommerce_{$this->id}_settings" === $option_data['id']
		) {
			$this->handle_uploaded_cert_file(
				$_FILES['woocommerce_' . $this->id . '_' . self::get_cert_info_by_type('public', 'filename' ) ] ?? array(),
				self::get_cert_info_by_type('public', 'filename' ),
				'live_certificate',
				array(
					'pem',
					'p12',
				)
			);
			$this->handle_uploaded_cert_file(
				$_FILES['woocommerce_' . $this->id . '_' . self::get_cert_info_by_type('private', 'filename' )] ?? array(),
				self::get_cert_info_by_type('private', 'filename' ),
				'live_private_certificate',
				array(
					'key',
					'p12',
				),
				true
			);

			$this->handle_uploaded_cert_file(
				$_FILES['woocommerce_' . $this->id . '_' . self::get_cert_info_by_type( 'sandbox_public', 'filename' ) ] ?? array(),
				self::get_cert_info_by_type('sandbox_public', 'filename' ),
				'sandbox_certificate',
				array(
					'pem',
					'p12',
				)
			);
			$this->handle_uploaded_cert_file(
				$_FILES['woocommerce_' . $this->id . '_' . self::get_cert_info_by_type( 'sandbox_private', 'filename' ) ] ?? array(),
				self::get_cert_info_by_type('sandbox_private', 'filename' ),
				'sandbox_private_certificate',
				array(
					'key',
					'p12',
				),
				true
			);
		}
	}

	/**
	 * Handles uploaded certificate files.
	 *
	 * @param array $file                  $_FILES array
	 * @param string $filename             Name to used for new file. Without extension.
	 * @param string $option_name          Name of option to save file path to.
	 * @param array $accepted_extensions   Accepted file extensions.
	 * @param bool $private                Whether the file is a private key.
	 *
	 * @return string                      Path to file, or empty string if no file was uploaded.
	 */
	private function handle_uploaded_cert_file( $file, $filename, $option_name, $accepted_extensions, $private = false ) {
		if ( empty( $file['name'] ) ) {
			if ( ! empty( $file['error'] ) && 4 !== $file['error'] ) {
				$this->add_error( __( 'File error', 'moreflo-checkout' ) ) . ': ' . $file['error'];
			} else {
				return '';
			}
		}

		$upload_dir = self::get_user_certs_path();

		if ( ! $upload_dir ) {
			$this->add_error( __( 'Upload directory error.', 'moreflo-checkout' ) );
			return '';
		}

		$file_extension = strtolower( pathinfo( $file['name'], PATHINFO_EXTENSION ) );

		if ( ! in_array( $file_extension, $accepted_extensions ) ) {
			$this->add_error( __( 'Certificate file extension incorrect', 'moreflo-checkout' ) );
			return '';
		}

		// max size 1MB
		if ( $file['size'] > 1000000 ) {
			$this->add_error( __( 'Certificate file is too large', 'moreflo-checkout' ) );
			return '';
		}

		if ( $private && 'p12' === $file_extension ) {
			return $this->handle_private_p12_file( $upload_dir, $file, $filename, $option_name );
		}

		$filename      = sanitize_file_name( $filename . '.' . $file_extension );
		$uploaded_file = $upload_dir . '/' . $filename;

		// if file exists, delete it
		if ( file_exists( $uploaded_file ) ) {
			unlink( $uploaded_file );
		}

		if ( move_uploaded_file( $file['tmp_name'], $uploaded_file ) ) {
			$this->settings[ $option_name ] = $uploaded_file;
			return $uploaded_file;
		}

		$this->add_error( __( 'Failed to store certificate file.', 'moreflo-checkout' ) );
		return '';
	}

	/**
	 * Handles uploaded private key files in p12 format.
	 *
	 * @param string $upload_dir   	 Upload directory.
	 * @param array $file            $_FILES array
	 * @param string $filename       Name to used for new file. Without extension.
	 * @param string $option_name    Name of option to save file path to.
	 *
	 * @return string                Path to file, or empty string if no file was uploaded.
	 */
	private function handle_private_p12_file( $upload_dir, $file, $filename, $option_name ) {
		$p12_content = file_get_contents( $file['tmp_name'] );

		if ( ! $p12_content ) {
			$this->add_error( __( 'Unable to read private .p12 file.', 'moreflo-checkout' ) );
			return '';
		}

		$is_p12_valid = openssl_pkcs12_read( $p12_content, $cert_info, "swish" );

		if ( ! $is_p12_valid ) {
			$this->add_error( __( 'Unable to read private .p12 file store.', 'moreflo-checkout' ) );
			return '';
		}

		if ( empty( $cert_info['pkey'] ) ) {
			$this->add_error( __( '.p12 file doesn\'t contain private key.', 'moreflo-checkout' ) );
			return '';
		}

		$filename      = sanitize_file_name( $filename . '.key' );
		$uploaded_file = $upload_dir . '/' . $filename;
		$upload_status = file_put_contents( $uploaded_file, $cert_info['pkey'] );

		if ( ! $upload_status ) {
			$this->add_error( __( 'Failed to store private key file.', 'moreflo-checkout' ) );
			return '';
		}

		$this->settings[ $option_name ] = $uploaded_file;
		return $uploaded_file;
	}

	/**
	 * Displays errors on settings page.
	 *
	 * @return void
	 */
	public function display_swish_errors() {
		global $current_section;
		if ( $current_section === $this->id ) {
			$this->display_errors();
		}
	}

	/**
	 * Renders certificate form modal.
	 *
	 * @return void
	 */
	public static function render_certificate_form_modal() {
		$email_address = get_option( 'admin_email' );
		$blog_url = get_bloginfo( 'url' );
		$domain = preg_replace( '#^https?://#i', '', $blog_url );
		?>
		<div id="nm-generate-cert-modal-container">
			<div id="nm-generate-cert-modal">
				<h3><?php esc_html_e( 'Information for certificate', 'moreflo-checkout' ); ?></h3>
				<p>
					<label for="nm-generate-new-cert-country"><?php esc_html_e( 'Country', 'moreflo-checkout' ); ?></label><br>
					<input type="text" id="nm-generate-new-cert-country" value="SE" readonly>
				</p>
				<p>
					<label for="nm-generate-new-cert-city"><?php esc_html_e( 'City', 'moreflo-checkout' ); ?></label><br>
					<input type="text" id="nm-generate-new-cert-city">
				</p>
				<p>
					<label for="nm-generate-new-cert-organization"><?php esc_html_e( 'Organization name', 'moreflo-checkout' ); ?></label><br>
					<input type="text" id="nm-generate-new-cert-organization">
				</p>
				<p>
					<label for="nm-generate-new-cert-common-name"><?php esc_html_e( 'Common name (domain)', 'moreflo-checkout' ); ?></label><br>
					<input type="text" id="nm-generate-new-cert-common-name" value="<?php echo esc_attr( $domain ); ?>">
				</p>
				<p>
					<label for="nm-generate-new-cert-email"><?php esc_html_e( 'Email', 'moreflo-checkout' ); ?></label><br>
					<input type="text" id="nm-generate-new-cert-email" value="<?php echo esc_attr( $email_address ); ?>">
				</p>
				<button id="nm-generate-new-cert"><?php esc_html_e('Generate','moreflo-checkout' ); ?></button>
				<button id="nm-generate-new-cert-cancel"><?php esc_html_e('Cancel','moreflo-checkout' ); ?></button>
			</div>
		</div>
		<?php
	}

	/**
	 * Enqueue admin scripts files.
	 *
	 * @return void
	 */
	public function admin_scripts() {

		wp_enqueue_style(
			'moreflo-checkout-swish-admin',
			plugins_url( 'css/moreflo_swish_settings.css', __FILE__ ),
			array(),
			filemtime( DIRPATH . 'css/moreflo_swish_settings.css' )
		);

		wp_enqueue_script(
			'moreflo-checkout-swish-admin',
			plugins_url( 'js/moreflo_swish_settings.js', __FILE__ ),
			array( 'jquery' ),
			filemtime( DIRPATH . 'js/moreflo_swish_settings.js' ),
			true
		);

		wp_localize_script(
			'moreflo-checkout-swish-admin',
			'moreflo_swish_settings',
			array(
				'ajax_url'                      => admin_url( 'admin-ajax.php' ),
				'nonce'                         => wp_create_nonce( 'moreflo_swish_settings' ),
				'abort_nonce'                   => wp_create_nonce( 'moreflo_swish_settings_abort_request' ),
				'generate_cert_nonce'           => wp_create_nonce( 'moreflo_swish_settings_generate_certificate' ),
				'connection_test_error_message' => __( 'An error occurred when running connection test, please try again.', 'moreflo-checkout' ),
				'cert_generation_error_message' => __( 'The following error occurred when generating certificate, did you fill in all the fields?', 'moreflo-checkout' ),
			)
		);
	}

   /**
    * You will need it if you want your custom credit card form, Step 4 is about it
    */
   public function payment_fields() {
    echo '<p class="form-row form-row-wide">
    <label for="' . esc_attr($this->id) . '_payer">' . __( 'Your cell phone number', 'moreflo-checkout' ) . '<span class="required">*</span></label>
    <input id="' . esc_attr($this->id) . '_payer" class="input-text ' . esc_attr($this->id) . '_payer" type="text" autocomplete="off" name="' . esc_attr($this->id) . '_payer" maxlength="17" /></p>';

   }

   function thankyou_scripts() {
       if ( ! is_checkout() && ! empty( is_wc_endpoint_url('order-received') ) ) {
           return;
       }

       wp_enqueue_style( 'mf-swish-pay', plugin_dir_url(__FILE__). 'css/moreflow_pay_swish.css', [], filemtime( plugin_dir_path(__FILE__). 'css/moreflow_pay_swish.css' ) );
   }

    function moreflo_thankyou($order_id)
    {
        $order = wc_get_order($order_id);
        if (is_object($order) && $order->get_payment_method() == 'swish_pay_gateway') {
            $logo = plugin_dir_url(__FILE__) . 'images/swish_horizontal_main_rgb.png';

            include('partials/swish_thank_you.php');
        }
    }

    public function local_certificate_info($swish_cert)
    {
        try {

            if (file_exists($swish_cert)) {
                $cert = file_get_contents($swish_cert);
                if ($cert_info = openssl_x509_parse($cert)) {
                    return (object) array(
                        'issuer' => $cert_info['issuer']['O'],
                        'common_name' => $cert_info['issuer']['CN'],
                        'valid_from' => $cert_info['validFrom_time_t'],
                        'valid_to' => $cert_info['validTo_time_t'],
                        'merchant_number' => $cert_info['subject']['CN'],
                    );
                }
            }

        } catch (Throwable $t) {
            if (method_exists($t, 'write_to_logs')) {
                $t->write_to_logs();
            } else {
                $this->logger->add( 'cert-info', print_r( $t, true ) );
            }

        }
        return false;
    }
   /*
     * Fields validation, more in Step 5
    */
   public function validate_fields() {

        $payer_raw = isset($_POST[esc_attr($this->id) . '_payer']) ? $_POST[esc_attr($this->id) . '_payer'] : '';
        $payer = preg_replace('/[^0-9]/', '', $payer_raw);
        $number_length = strlen($payer);
        if ($number_length == 0) {
            wc_add_notice(__('Swish number missing', 'moreflo-checkout'), 'error');
            return false;
        }
        if ($payer[0] == '0') {
            $payer = '46' . substr($payer, 1);
            $number_length = strlen($payer);
        }
        if ($number_length < 8) {
            wc_add_notice(__('Swish number must be at least 8 characters long', 'moreflo-checkout'), 'error');
            return false;
        }
        if ($number_length > 15) {
            wc_add_notice(__('Swish number can be maximum 15 characters long', 'moreflo-checkout'), 'error');
            return false;
        }
        $this->payer_number = $payer;
        return true;

   }

	/**
	 * Test Swish connection. Response is sent as JSON, then WP dies.
	 *
	 * @return void
	 */
   public function connection_test() {

	   if (
		   empty( $this->swish_connection_test_enabled ) ||
		   'no' === $this->swish_connection_test_enabled
	   ) {
		   wp_send_json_success( '' );
	   }

	   $error_message = false;

	   if ( empty( $this->api_end_URL ) ) {
		   $error_message = __( 'Please enter API URL', 'moreflo-checkout' );
	   }

	   if ( empty( $this->payee_number ) ) {
		   $error_message = __( 'Please enter Merchant Swish number', 'moreflo-checkout' );
	   }

	   if ( empty( $this->public_certificate ) ) {
		   $error_message = __( 'Please enter Swish certificate', 'moreflo-checkout' );
	   }

	   if ( empty( $this->root_certificate ) ) {
		   $error_message = __( 'Swish Root certificate missing', 'moreflo-checkout' );
	   }

	   if ( 'prod' === $this->swish_environment && empty( $this->swish_connection_test_payer ) ) {
		   $error_message = __( 'When testing the production environment you have to enter a real payer number.', 'moreflo-checkout' );
	   }

	   if ( ! $error_message ) {
		   $test_payer = '46701740635';

		   if ( ! empty( $this->swish_connection_test_payer ) ) {
			   $test_payer = $this->swish_connection_test_payer;
			   if ($test_payer[0] == '0') {
				   $test_payer = '46' . substr($test_payer, 1);
			   }
		   }

		   $swishAPI = new Moreflo_swish_api(
				   $this->api_end_URL,
				   $this->payee_number,
				   $this->root_certificate,
				   $this->public_certificate,
				   $this->private_certificate,
				   $this->private_key ?? '',
				   $this->log_enabled ?? ''
		   );

		   $args = array(
			   'wc-api' => 'swish_pay_gateway_connection_test',
			   'event' => 'payment',
			   'nonce' => 'test_payment',
		   );

		   $callback_url = add_query_arg($args, site_url('/'));

		   $response = $swishAPI->apiRequest(
			   array(
		            'payeePaymentReference' => '0123456789',
		            'callbackUrl' => $callback_url,
		            'payerAlias' => $test_payer,
		            'payeeAlias' => $this->payee_number,
		            'amount' => '1',
		            'currency' => 'SEK',
		            'message' => 'Northmill Test Product',
		        ),
			   'paymentrequests'
		   );

		   if ( $response['status'] ) {


			   preg_match( '/Location:.*paymentrequests\/(.*?)\n/', $response['content'] ?? '', $matches );
			   $request_id = trim( array_pop( $matches ) );

			   wp_send_json_success(
					array(
						'message'    => '<strong style="color: green;">' . __( 'Connection test successful', 'moreflo-checkout' ) . '</strong>',
						'payment_id' => $request_id,
					)
			   );
		   } else {
			   $error_message = $response['response_technical'] ?? '';
		   }
	   } else {
		   $error_message = __( 'Invalid connection settings', 'moreflo-checkout' ) . ': ' . $error_message;
	   }

	   wp_send_json_error( '<strong style="color: red;">' . $error_message . '</strong>' );
   }

	/**
	 * Function called by AJAX to test the connection to Swish.
	 *
	 * @return void
	 */
	public static function run_connection_test() {
		if ( ! wp_verify_nonce( $_POST['nonce'], 'moreflo_swish_settings' ) ) {
			wp_send_json_error( 'Invalid nonce' );
		}

		// Get this gateway
		$gateways = WC()->payment_gateways->payment_gateways();

		$gateway = $gateways['swish_pay_gateway'] ?? false;

		if ( ! $gateway ) {
			wp_send_json_error( __( 'Swish payment gateway not found', 'moreflo-checkout' ) );
		}

		$gateway->connection_test();
	}

	/**
	 * Abort Swish payment request.
	 *
	 * @param string $request_id Swish payment request ID.
	 *
	 * @return array|bool Response from Swish API. Or false if called incorrectly.
	 */
	public function abort_request( $request_id ) {
		if ( 'mss' === $this->swish_environment ) { // MSS requests cannot be aborted.
			return false;
		}

		if ( empty( $request_id ) ) {
			return false;
		}
		$log_enabled = $this->log_enabled ?? '';

		$swishAPI = new Moreflo_swish_api(
				$this->api_end_URL,
				$this->payee_number,
				$this->root_certificate,
				$this->public_certificate,
				$this->private_certificate,
				$this->private_key,
				$log_enabled
		);
		return $swishAPI->apiRequest(
			array(
				array(
					'op'    => 'replace',
					'path'  => '/status',
					'value' => 'cancelled',
				),
			),
		   "paymentrequests/$request_id",
		   'PATCH'
		);
	}

	/**
	 * Function called by AJAX to abort Swish payment request.
	 *
	 * @return void
	 */
	public static function abort_connection_test_request() {
		if ( ! wp_verify_nonce( $_POST['nonce'], 'moreflo_swish_settings_abort_request' ) ) {
			wp_send_json_error( 'Invalid nonce' );
		}

		// Get this gateway
		$gateways = WC()->payment_gateways->payment_gateways();
		$gateway  = $gateways['swish_pay_gateway'] ?? false;

		if ( ! $gateway ) {
			wp_send_json_error( __( 'Swish payment gateway not found', 'moreflo-checkout' ) );
		}

		$request_id = sanitize_text_field( stripslashes( $_POST['payment_id'] ?? '' ) );

		$response = $gateway->abort_request( $request_id );

		if ( $response && $response['status'] ) {
			wp_send_json_success( __( 'Request canceled', 'moreflo-checkout' ) );
		} else {
			wp_send_json_error( __( 'Could not cancel request', 'moreflo-checkout' ) );
		}
	}

	/**
	 * Generate new certificate request file, and private key.
	 *
	 * @return array|bool Array with error messages, or true on success.
	 */
	public function do_generate_new_certificate() {
		// for SSL server certificates the commonName is the domain name to be secured
		// for S/MIME email certificates the commonName is the owner of the email address
		// location and identification fields refer to the owner of domain or email subject to be secured
		$city = wp_unslash( sanitize_text_field( $_POST['city'] ?? '' ) );
		$dn = array(
			"countryName"            => "SE",
			"stateOrProvinceName"    => $city,
			"localityName"           => $city,
			"organizationName"       => wp_unslash( sanitize_text_field( $_POST['organization'] ?? '' ) ),
			"organizationalUnitName" => "Web",
			"commonName"             => wp_unslash( sanitize_text_field( $_POST['common_name'] ?? '' ) ),
			"emailAddress"           => wp_unslash( sanitize_text_field( $_POST['email'] ?? '' ) ),
		);

		// Generate a new private (and public) key pair
		$private_key = openssl_pkey_new(
			array(
				"private_key_bits" => 4096,
			    "private_key_type" => OPENSSL_KEYTYPE_RSA,
			)
		);

		// Generate a certificate signing request
		$csr = openssl_csr_new(
			$dn,
			$private_key,
			array( 'digest_alg' => 'sha256' )
		);

		// Generate a self-signed cert, valid for 365 days
		$x509 = openssl_csr_sign(
			$csr,
			null,
			$private_key,
			365,
			array( 'digest_alg' => 'sha256' )
		);

		$certificate_private_key = wp_generate_password( 12, false );

		// Save your private key, CSR and self-signed cert for later use
		openssl_csr_export( $csr, $csr_out );
		openssl_pkey_export( $private_key, $pkey_out, $certificate_private_key );

		$errors = array();

		/**
		 * Openssl error messages to ignore.
		 * Error messages will be checked if they contain any of the strings in this array.
		 * The string doesn't have to be the full error message, it can be a part of it.
		 */
		$errors_to_ignore = array(
			'NCONF_get_string:no value',
		);

		// Show any errors that occurred here
		while ( ( $e = openssl_error_string() ) !== false ) {
			$contains_error_to_ignore = false;
			foreach ( $errors_to_ignore as $error_to_ignore ) {
				if ( strpos( $e, $error_to_ignore ) !== false ) {
					$contains_error_to_ignore = true;
					break;
				}
			}

			if ( ! $contains_error_to_ignore ) {
				$errors[] = $e;
			}
		}

		if ( ! empty( $errors ) ) {
			return $errors;
		}

		$stored_csr_file   = self::get_cert_info_by_type( 'csr', 'filepath' );
		$csr_upload_status = file_put_contents( $stored_csr_file, $csr_out );

		if ( ! $csr_upload_status ) {
			$errors[] = __( 'Failed to store CSR file.', 'moreflo-checkout' );
			return $errors;
		}

		$stored_private_certificate_file = self::get_cert_info_by_type('private', 'filepath' ) . 'key';
		$private_upload_status           = file_put_contents( $stored_private_certificate_file, $pkey_out );

		if ( ! $private_upload_status ) {
			$errors[] = __( 'Failed to store private key file.', 'moreflo-checkout' );
			return $errors;
		}

		$this->update_option( 'live_certificate', '' ); // Clear public certificate, since it has to be uploaded manually.
		$this->update_option( 'live_private_certificate', $stored_private_certificate_file );
		$this->update_option( 'live_csr_file', $stored_csr_file );
		$this->update_option( 'live_private_key', $certificate_private_key );
		return true;
	}

	/**
	 * Function called by AJAX to generate new certificate request.
	 *
	 * @return void
	 */
	public static function generate_new_certificate() {
		if ( ! wp_verify_nonce( $_POST['nonce'], 'moreflo_swish_settings_generate_certificate' ) ) {
			wp_send_json_error( 'Invalid nonce' );
		}

		// Get this gateway
		$gateways = WC()->payment_gateways->payment_gateways();
		$gateway  = $gateways['swish_pay_gateway'] ?? false;

		if ( ! $gateway ) {
			wp_send_json_error( __( 'Swish payment gateway not found', 'moreflo-checkout' ) );
		}

		$result = $gateway->do_generate_new_certificate();

		if ( is_array( $result ) ) {
			wp_send_json_error( $result );
		} else {
			wp_send_json_success(
				array(
					'private_key' => $gateway->get_option( 'live_private_key' ),
					'message'     => __( 'Certificate generated', 'moreflo-checkout' ),
					'download'    => get_admin_url( null, 'admin-post.php?action=download_cert&type=csr' ),
				)
			);
		}
	}

	/**
	 * Function called /wp-admin/admin-post.php to download certificate files.
	 *
	 * @return void
	 */
	public static function download_cert() {
		$type = stripslashes( sanitize_text_field( $_GET['type'] ?? '' ) );
		$filepath = false;

		switch ( $type ) {
			case 'csr':
				$filepath = self::get_cert_info_by_type( 'csr', 'filepath' );
				break;

			case 'private':
			case 'public':
			case 'sandbox_private':
			case 'sandbox_public':
				$filepath = self::get_stored_cert_file_path( $type );
				break;
		}

		if ( ! $filepath ) {
			wp_die( __( 'Invalid certificate type', 'moreflo-checkout' ) );
		}

		$filename = basename( $filepath );
		$filetype = filetype( $filepath );
		header( 'Content-Disposition: attachment; filename=' . $filename );
		header( "Content-type: " . $filetype );
		ob_clean();
		flush();
		readfile( $filepath );
		exit;
	}

	/**
	 * Get certificate file path.
	 *
	 * @param string $type Type of certificate.
	 * @return string
	 */
	public static function get_stored_cert_file_path( $type ) {

		// Get this gateway
		$gateways = WC()->payment_gateways->payment_gateways();
		$gateway  = $gateways['swish_pay_gateway'] ?? false;

		if ( ! $gateway ) {
			wp_send_json_error( __( 'Swish payment gateway not found', 'moreflo-checkout' ) );
		}

		$cert_info = self::get_cert_info_by_type( $type );

		return ! empty( $cert_info['option_name'] ) ? $gateway->get_option( $cert_info['option_name'] ) : false;
	}

    /**
     * Begins processing payment. Will redirect user to Swish waiting page on success, else add wc notice and return empty.
     *
     * @param int $order_id
     */
    public function process_payment( $order_id ) {
        $order = new WC_Order( $order_id );

        $this->write_Woo_logs('process_payment');

        $swishClient = new Moreflo_swish_client();
        $customerObj = $swishClient->preparePaymentClient($order, $this->payer_number, $this->payee_number);

        $this->write_Woo_logs($customerObj);

        $this->write_Woo_logs('Endpoint: '. $this->api_end_URL);

        $swishAPI = new Moreflo_swish_api(
				$this->api_end_URL,
				$this->payee_number,
				$this->root_certificate,
				$this->public_certificate,
				$this->private_certificate,
				$this->private_key,
				$this->log_enabled
        );
        $payment = $swishAPI->apiRequest($customerObj, 'paymentrequests');

        $this->write_Woo_logs($payment);
        $this->write_Woo_logs('process_payment End');

        if ($payment['status']) {
            update_post_meta($order_id, '_swish_payment_location', $payment['response']);
            update_post_meta($order_id, '_swish_payment_status', 'WAITING');

            // Mark as pending (we're awaiting the Swish confirmation)
            $order->add_order_note(__('Payment initiated', 'moreflo-checkout'));
            $order->update_status('wc-pending', __( 'Awaiting Swish payment confirmation', 'moreflo-checkout' ));

            // Redirect user to Swish payment waiting page
            return array(
                'result' => 'success',
                'redirect' => add_query_arg(
                    [
                        'event_type' => 'proccess',
                        'order_id' => $order_id,
                        'order_key' => $order->get_order_key('edit'),
                    ],
                    WC()->api_request_url( 'swish_pay_gateway_wait' )
                ),
            );
        }

        wc_add_notice( 'Error : '.$payment['response'] , 'error' );
    }

    /**
     * Process refunds
     *
     * @param  int $order_id
     * @param  float $amount
     * @param  string $reason
     * @return bool|WP_Error
     */
    public function process_refund($order_id, $amount = null, $reason = '')
    {
        $this->write_Woo_logs('process_refund');
        $this->write_Woo_logs($order_id);

        try {

            $order = wc_get_order( $order_id );

            $transaction_id = $order->get_transaction_id();

            if (!$transaction_id) {
                $this->write_Woo_logs(sprintf("No transaction ID for order: %s", $order_id));
            }

            $swishClient = new Moreflo_swish_client();
            $refundcustomerObj = $swishClient->prepareRefundPaymentClient($transaction_id, $order, $amount, $reason, $this->payee_number);

            $this->write_Woo_logs($refundcustomerObj);

            $swishAPI = new Moreflo_swish_api(
					$this->api_end_URL,
					$this->payee_number,
					$this->root_certificate,
					$this->public_certificate,
					$this->private_certificate,
					$this->private_key,
					$this->log_enabled
            );
            $payment = $swishAPI->apiRequest($refundcustomerObj, 'refunds');

            $this->write_Woo_logs($payment);

            $this->write_Woo_logs('process_refund End');
            if($payment['status']){
                return true;
            }else{
                $this->write_Woo_logs(sprintf('Received error as (%s) for refund order: %s , from Swish', $payment['response'], $order_id));
                return false;
            }

        } catch (Exception $e) {
            throw new Exception($e);
        }
    }

    public function swish_wait_callback_response() {
        if ( ! isset( $_GET['order_id'] ) || ! isset( $_GET['order_key'] ) ) {
            echo '<a href="'. esc_attr( wc_get_cart_url() ) .'">'. esc_html__( 'Something went wrong, please go back and try again.', 'moreflo-checkout' );
            exit;
        }
        if ( ! ( isset( $_GET['event_type'] ) && 'proccess' == $_GET['event_type'] ) ) {
            echo '<a href="'. esc_attr( wc_get_cart_url() ) .'">'. esc_html__( 'Something went wrong, please go back and try again.', 'moreflo-checkout' );
            exit;
        }
        $order_id = wc_get_order_id_by_order_key($_GET['order_key']);
        if ( $order_id != $_GET['order_id'] ) {
            echo '<a href="'. esc_attr( wc_get_cart_url() ) .'">'. esc_html__( 'Something went wrong, please go back and try again.', 'moreflo-checkout' );
            exit;
        }

        $order = wc_get_order( $order_id );

        $css_link = add_query_arg([
            'v' => filemtime( plugin_dir_path(__FILE__). 'css/moreflow_pay_swish.css' ),
        ], plugin_dir_url(__FILE__). 'css/moreflow_pay_swish.css');
        $js_links = [
            add_query_arg([
                'v' => filemtime( ABSPATH . WPINC . '/js/jquery/jquery.js' ),
            ], '/wp-includes/js/jquery/jquery.js'),
            add_query_arg([
                'v' => filemtime( plugin_dir_path( __FILE__ ) . 'js/moreflo_swish_pay.js' ),
            ], plugin_dir_url( __FILE__ ) . 'js/moreflo_swish_pay.js'),
        ];
        $js_local = [
            'ajaxurl' => admin_url( 'admin-ajax.php' ),
            'nonce' => wp_create_nonce('moreflo_swish_pay'),
            'redirect_success' => $order->get_checkout_order_received_url(),
            'redirect_cancelled' => apply_filters( 'moreflo_checkout_swish_pay_redirect_cancelled', wc_get_checkout_url() ),
            'redirect_cancel' => $order->get_cancel_order_url_raw(),
        ];
        $logo = plugin_dir_url(__FILE__) . 'images/swish_horizontal_main_rgb.png';

        $swish_wait_url = add_query_arg(
            [
                'event_type' => 'proccess',
                'order_id' => $order->get_id(),
                'order_key' => $order->get_order_key( 'edit' ),
            ],
            WC()->api_request_url( 'swish_pay_gateway_wait' )
        );

        $swish_app = 'swish://';

		$swish_environment = $this->swish_environment;

        include('partials/swish_awaiting_payment.php');
        exit;
    }

   /*
    * Callback methods
    */
    public function swish_callback_response() {
        $this->write_Woo_logs('swish_callback_response');
        if($_REQUEST['event']){
            $this->write_Woo_logs($_REQUEST);

            $event_type = $_REQUEST['event'];
            $this->write_Woo_logs($event_type);

            if($event_type == 'payment' || $event_type == 'refund'){
                $nonce_verify = $_REQUEST['nonce'];
                $order_id = $_REQUEST['order_id'];

                $order = wc_get_order( $order_id );

                $swishClient = new Moreflo_swish_client();
                $verifiedCustomer = $swishClient->verify_callback($nonce_verify, $order_id, $this->payee_number);

                $this->write_Woo_logs($verifiedCustomer);

                if($verifiedCustomer){
                    $request_body = file_get_contents("php://input");
                    $json = json_decode($request_body);

                    if (empty($request_body)) {
                        $this->write_Woo_logs(sprintf('Request body missing for order %s', $order_id));
                        wp_die();
                    }else{
                        $this->write_Woo_logs(sprintf('Request body details order id %s details of response body is %s', $order_id, $request_body));
                    }

                    switch ($json->status) {
                        case 'DECLINED':
                            if ( $order->get_payment_method() == $this->id &&
                                in_array( $order->get_status(), [ 'pending', 'cancelled' ] )
                            ) {
                                $order->add_order_note( __('MoreFlo - Payment declined by user', 'moreflo-checkout') );
                                $order->update_status('wc-cancelled');
                                $this->write_Woo_logs('Payment declined by user');
                            } else {
                                $order->add_order_note( __('MoreFlo Swish - Payment declined by user but order status cannot be updated (not pending swish payment)', 'moreflo-checkout') );
                                $this->write_Woo_logs('Payment declined by user (Failed to set DECLINED status on '. $order_id .')');
                            }
                            break;
                        case 'ERROR':
                            $swishError = new Moreflo_swish_errors();
                            $error_details = $swishError->error_code($json->errorCode);
                            $this->write_Woo_logs('Received ERROR : '.$error_details.', from Swish on order '. $order_id);

                            if ( $order->get_payment_method() == $this->id &&
                                in_array( $order->get_status(), [ 'pending', 'cancelled' ] )
                            ) {
                                $order->add_order_note($json->errorCode.' - '.$error_details );
                                $order->update_status('wc-failed');
                            } else {
                                $order->add_order_note( __('MoreFlo Swish - Payment error but order status cannot be updated (not pending swish payment)', 'moreflo-checkout') );
                                $this->write_Woo_logs('Payment error (Failed to set ERROR status on order '. $order_id .')');
                            }
                            break;
                        case 'DEBITED':
                            update_post_meta($order_id, '_swish_refund_debited', true);
                            $order->update_status( 'wc-pending' );
                            $order->add_order_note( sprintf(__('MoreFlo - Merchant account debited - %s ID: %s', 'moreflo-checkout'),'Transaction', $json->id) );
                            $this->write_Woo_logs('Got DEBITED from Swish');
                            break;
                        case 'PAID':
                            if($event_type != 'refund'){
                                $order->add_order_note( sprintf(__('MoreFlo - Payment confirmed - %s ID: %s', 'moreflo-checkout'),'Transaction', $json->id) );
                                $order->set_transaction_id($json->paymentReference);
                                $order->payment_complete();
                                $this->write_Woo_logs('Got PAID from Swish');
                            }
                            if((get_post_meta($order_id, '_swish_refund_debited', true)) && ($event_type == 'refund')){
                                $order->update_status( 'wc-refunded' );
                                $this->write_Woo_logs('Successfully credited to customer account');
                                $order->add_order_note( sprintf(__('MoreFlo - Successfully credited to customer account - %s ID: %s', 'moreflo-checkout'),'Transaction', $json->id) );
                            }
                            break;
                    }
                    $order->save();

                    $returnResult['success'] = true;
                    $returnResult['successCode'] = "200";

                    wp_send_json( $returnResult, $returnResult['successCode'] );//Send received notification to server

                }else{
                    $this->write_Woo_logs('Not a valid customer');
                }
                $this->write_Woo_logs('swish_callback_response End');
            }
        }
    }
	public function swish_connection_test_callback_response() {
		$this->write_Woo_logs('swish_connection_test_callback_response');
		$this->write_Woo_logs($_REQUEST);
		$this->write_Woo_logs('swish_connection_test_callback_response End');
	}

	/**
	 * Allows for testing swish payments locally. Requires constant WP_ENVIRONMENT_TYPE be set to 'local' and Swish payment gateway environment to be sandbox or mss.
	 *       Trigger at:
				/?wc-api=swish_pay_gateway&event=payment&status=paid&order_id=42
	 *       Replace order_id. Parameters event and status can also be changed.
	 * @return void
	 * @throws WC_Data_Exception
	 */
	public function swish_callback_response_local_dev() {
		if (
			( 'sandbox' !== $this->swish_environment && 'mss' !== $this->swish_environment ) ||
			'local' !== wp_get_environment_type() )
		{
			$this->write_Woo_logs( 'Aborting attempt for local Swish callback, order# ' . ( $_REQUEST['order_id'] ?? '' ) . '. WP ENV: ' .  wp_get_environment_type() . ' Swish env: ' . $this->swish_environment );
			return;
		}

		$this->write_Woo_logs('swish_callback_response_LOCAL_DEV');

		if ( $_REQUEST['event'] ) {
			$this->write_Woo_logs( $_REQUEST );
			$event_type = $_REQUEST['event'];

			if ( $event_type == 'payment' || $event_type == 'refund' ) {
				$order_id = $_REQUEST['order_id'];
				$order    = wc_get_order( $order_id );

				switch ( strtoupper( $_REQUEST['status'] ) ) {
					case 'DECLINED':
						if ( $order->get_payment_method() == $this->id &&
						     in_array( $order->get_status(), [ 'pending', 'cancelled' ] )
						) {
							$order->add_order_note( __( 'MoreFlo - Payment declined by user', 'moreflo-checkout' ) );
							$order->update_status( 'wc-cancelled' );
							$this->write_Woo_logs( 'Payment declined by user' );
						} else {
							$order->add_order_note( __( 'MoreFlo Swish - Payment declined by user but order status cannot be updated (not pending swish payment)', 'moreflo-checkout' ) );
							$this->write_Woo_logs( 'Payment declined by user (Failed to set DECLINED status on ' . $order_id  . ')' );
						}
						break;

					case 'ERROR':
						$this->write_Woo_logs( 'Received ERROR : DEBUG ERROR, from Swish on order ' . $order_id );

						if ( $order->get_payment_method() == $this->id &&
						     in_array( $order->get_status(), [ 'pending', 'cancelled' ] )
						) {
							$order->add_order_note( 'SWISH DEBUG ERROR' );
							$order->update_status( 'wc-failed' );
						} else {
							$order->add_order_note( __( 'MoreFlo Swish - Payment error but order status cannot be updated (not pending swish payment)', 'moreflo-checkout' ) );
							$this->write_Woo_logs( 'Payment error (Failed to set ERROR status on order ' . $order_id . ')' );
						}
						break;

					case 'DEBITED':
						update_post_meta( $order_id, '_swish_refund_debited', true );
						$order->update_status( 'wc-pending' );
						$order->add_order_note( sprintf( __( 'MoreFlo - Merchant account debited - %s ID: %s', 'moreflo-checkout' ),'Transaction', 'DEBUG 42' ) );
						$this->write_Woo_logs( 'Got DEBITED from Swish' );
						break;

					case 'PAID':
						if($event_type != 'refund'){
							$order->add_order_note( sprintf( __( 'MoreFlo - Payment confirmed - %s ID: %s', 'moreflo-checkout' ),'Transaction', 'DEBUG 42' ) );
							$order->set_transaction_id( 'DEBUG 42' );
							$order->payment_complete();
							$this->write_Woo_logs( 'Got PAID from Swish' );
						}
						if ( ( get_post_meta($order_id, '_swish_refund_debited', true ) ) && ( $event_type == 'refund' ) ) {
							$order->update_status( 'wc-refunded' );
							$this->write_Woo_logs('Successfully credited to customer account');
							$order->add_order_note( sprintf( __( 'MoreFlo - Successfully credited to customer account - %s ID: %s', 'moreflo-checkout' ),'Transaction', 'DEBUG 42' ) );
						}
						break;
				}
				$order->save();

				$returnResult['success'] = true;
				$returnResult['successCode'] = "200";

				echo '<pre>';
				print_r( $returnResult );
				echo '</pre>';

				$this->write_Woo_logs( 'swish_callback_response_LOCAL_DEV End' );
			}
		}
		wp_die( 'Swish local dev callback finished' );
	}

    public function write_Woo_logs($details)
    {
        if ( $this->log_enabled == 'yes' ) {
            $this->logger->debug(wc_print_r($details, true), $this->logger_context);
        }
    }

    public static function return_swish_pay_status()
    {
        if (!wp_verify_nonce($_POST['valid'], 'moreflo_swish_pay')) {
            wp_send_json('not verified');
            wp_die();
        }

        preg_match('/\=(wc_order_[^&]*)/', $_POST['url'], $key);

        if ($order_id = wc_get_order_id_by_order_key($key[1])) {

            $order = wc_get_order( $order_id );

            $status = $order->get_status();
            $status_translation = [
                'cancelled' => __( 'cancelled', 'moreflo-checkout' ),
                'pending' => __( 'pending', 'moreflo-checkout' ),
            ];
            $order_status_array = array( 'processing', 'completed', 'cancelled', 'failed' );
            $order_status_msg = array(
                'processing' => __("Thank you, You have successfully approved your payment", 'moreflo-checkout'),
                'completed' => __("Thank you, You have successfully approved your payment", 'moreflo-checkout'),
                'cancelled' => sprintf(__("Thank you, You have %s your payment", 'moreflo-checkout'), isset($status_translation[$status]) ? $status_translation[$status] : $status),
                'failed' => __("There were errors on you payment, Please contact shop for details", 'moreflo-checkout'));

            if(in_array($status, $order_status_array)){
                $response = array(
                    'error' => false,
                    'status' => $status,
                    'message' => $order_status_msg[$status],
                );
            }else{
                $response = array(
                    'error' => true,
                );
            }

        } else {

            $response = array(
                'error' => true,
                'message' => __('Error on recognize you order, please contact your shop', 'moreflo-checkout'),
            );

        }

        wp_send_json($response);

        wp_die();
    }
}
